import vip_ivp as vip
import numpy as np


def test_source_expression():
    lambda_fun = lambda t: t

    def fun(t):
        return t

    constant = vip.temporal(5)
    anonymous = vip.temporal(lambda t: t)
    lambda_source = vip.temporal(lambda_fun)
    fun_source = vip.temporal(fun)

    dict_source = vip.temporal({"a": 5, "b": 4})
    list_source = vip.temporal([1, 2, 3])
    array_source = vip.temporal(np.linspace(0, 10, 11))

    # Scalars
    assert constant._expression == str(5)
    assert anonymous._expression == "lambda t: t"
    assert lambda_source._expression == "lambda t: t"  # Assigned lambda should be identical
    assert fun_source._expression == "fun(t)"
    # Collection
    assert repr(dict_source) == str({"a": 5, "b": 4})  # Convert to string
    assert repr(list_source) == str([1, 2, 3])  # Convert list to string
    assert repr(array_source) == str(np.linspace(0, 10, 11))  # Convert array to list then string
    # Collection elements
    assert dict_source["a"]._expression == 'dict_source[a]'
    assert list_source[1]._expression == "list_source[1]"
    assert array_source[2]._expression == "array_source[2]"


def test_operations():
    a = vip.temporal(0)
    b = vip.temporal(1)
    c = vip.temporal(2)
    d = vip.temporal(3)

    # Testing addition
    ab = a + b
    assert ab.expression == "a + b"

    # Testing subtraction
    ac = a - c
    assert ac.expression == "a - c"

    # Testing multiplication
    ad = a * d
    assert ad.expression == "a * d"

    # Testing division
    ab_div = a / b
    assert ab_div.expression == "a / b"

    # Testing power
    ab_pow = a ** b
    assert ab_pow.expression == "a ** b"

    # Testing modulo
    ab_mod = a % b
    assert ab_mod.expression == "a % b"

    # Testing mixed operations
    e = a + b * c
    assert e.expression == "a + b * c"

    f = (a + b) * c
    assert f.expression == "(a + b) * c"

    g = a * (b + c)
    assert g.expression == "a * (b + c)"

    h = (a + b) ** 2
    assert h.expression == "(a + b) ** 2"

    i = a * b + c
    assert i.expression == "a * b + c"

    # Check more complex combinations
    j = (a + b) * (c + d)
    assert j.expression == "(a + b) * (c + d)"

    k = a + b + c + d
    assert k.expression == "a + b + c + d"

    l = (a + b) * c + d
    assert l.expression == "(a + b) * c + d"

    n = a + b * (c + d)
    assert n.expression == "a + b * (c + d)"

    # Testing more complicated expressions involving addition and subtraction
    p = a + b - c + d
    assert p.expression == "a + b - c + d"

    # Check multiple chained operations
    r = a + b + c + d
    assert r.expression == "a + b + c + d"

    # Check with other operations mixed in
    s = a + b * c + d
    assert s.expression == "a + b * c + d"

    t = np.sin(a, where=b > 1)
    assert t.expression == "sin(a, where=b > 1)"

    u = -(a + b)
    assert u.expression == "- (a + b)"

    v = a - (b + c)
    assert v.expression == "a - (b + c)"


def test_diff_equation():
    ddx = vip.loop_node()
    dx = vip.integrate(ddx, 0)
    x = vip.integrate(dx, 0)
    ddx.loop_into(2 * x - 4 * dx)

    assert dx.expression == "#INTEGRATE ddx"
    assert x.expression == "#INTEGRATE dx"
    assert ddx.expression == "2 * x - 4 * dx"


def test_transformations():
    a = vip.temporal(5)

    da = a.derivative()
    ia = vip.integrate(a, 0)
    delay_a = a.delayed(1)

    def foo(bar):
        return bar + 1

    foo_a = vip.f(foo)(a)

    assert da.expression == "#D/DT a"
    assert ia.expression == "#INTEGRATE a"
    assert delay_a.expression == "#DELAY(1) a"
    assert foo_a.expression == "foo(a)"


def test_trigger():
    a = vip.temporal(1)
    b = vip.integrate(5, 0)

    cond = a < 5
    other_cond = a > 2

    cross_constant = a.crosses(10)
    cross_temporalvar = a.crosses(b)
    cross_cond = cond.crosses(True)
    cross_cond_temporalvar = cond.crosses(other_cond)

    assert repr(cross_constant) == "#CROSSING_BETWEEN a AND 10"
    assert repr(cross_temporalvar) == "#CROSSING_BETWEEN a AND b"
    assert repr(cross_cond) == "#CROSSING_BETWEEN cond AND True"
    assert repr(cross_cond_temporalvar) == "#CROSSING_BETWEEN cond AND other_cond"

    vip.solve(10)


def test_events():
    a = vip.temporal(1)
    ia = vip.integrate(5, 0)

    trigger = ia.crosses(6)

    vip.solve(10)

    assert repr(ia.reset_on(trigger, 0)) == "Event(On trigger, Reset ia to 0)"
    assert repr(vip.terminate_on(trigger)) == "Event(On trigger, Terminate simulation)"



import numpy as np

from vip_ivp.utils import check_if_vectorized, vectorize_source


def test_check_if_vectorized():
    res1 = check_if_vectorized(scalar_only_fun)
    res2 = check_if_vectorized(constant_fun)
    res3 = check_if_vectorized(scalar_mode_fun)
    res4 = check_if_vectorized(vectorized_function)

    assert res1[0] == False
    assert res2[0] == False
    assert res3[0] == True and res3[1] == True
    assert res4[0] == True and res4[1] == False
    print(res3)


def test_vectorization():
    scalar_input = 1
    array_input = np.linspace(0, 5, 6)

    # Test 0D functions
    fun_list = [scalar_only_fun, scalar_mode_fun, vectorized_function]
    for fun in fun_list:
        assert vectorize_source(fun)(scalar_input) == scalar_input
        assert np.array_equal(vectorize_source(fun)(array_input), array_input)
    # Test constant function
    assert vectorize_source(constant_fun)(scalar_input) == 1
    assert np.array_equal(vectorize_source(constant_fun)(array_input), np.ones_like(array_input))

    # Test 1D functions
    assert np.array_equal(vectorize_source(constant_arr)(scalar_input), np.ones((2,3)))
    assert np.array_equal(vectorize_source(constant_arr)(array_input), np.ones((2,3,len(array_input))))



def scalar_only_fun(t):
    return max(t, 0)


def scalar_mode_fun(t):
    if np.isscalar(t):
        return t
    else:
        return t


def vectorized_function(t):
    if np.isscalar(t):
        return np.array([t])
    return np.array(t)


def constant_fun(t):
    return 1

def constant_arr(t):
    return np.ones((2,3))



import inspect
import operator
import types
import warnings
from typing import Any, Generator, Callable

import numpy as np


def vectorize_source(fun: Callable) -> Callable:
    accept_arrays, has_scalar_mode, is_constant = check_if_vectorized(fun)

    def vectorized_wrapper(t):
        if np.isscalar(t):
            return fun(t)
        else:
            return np.array([fun(ti) for ti in t])

    def handle_scalar_wrapper(t):
        output = fun(t)
        if np.isscalar(t):
            return output[0]
        return output

    def constant_wrapper(t):
        if np.isscalar(t):
            return fun(t)
        else:
            value = fun(0)
            if np.isscalar(value):
                return np.full(len(t), value)
            else:
                return np.moveaxis(np.broadcast_to(value, (len(t),) + value.shape), 0, -1)

    if is_constant:
        return constant_wrapper
    elif accept_arrays and has_scalar_mode:
        return fun
    elif not has_scalar_mode:
        return handle_scalar_wrapper
    else:
        print(f"Warning: the function '{fun.__name__}' is not vectorizable.")
        return vectorized_wrapper


def check_if_vectorized(fun) -> (bool, bool, bool):
    accept_arrays = True
    has_scalar_mode = True
    is_constant = False

    # Test with scalar
    scalar_output = fun(0)

    # Find a t length that does not match the len of a dimension of a scalar output
    array_len = 3
    if not np.isscalar(scalar_output):
        while array_len in scalar_output.shape:
            array_len += 1

    # Test with array input
    array_input = np.zeros(array_len)
    try:
        array_output = fun(array_input)
        if np.isscalar(array_output) or not np.isscalar(scalar_output) and array_output.shape == scalar_output.shape:
            is_constant = True
            raise ValueError("The array output does not create a vector")
        scalar_ndim = 0 if np.isscalar(scalar_output) else np.asarray(scalar_output).ndim
        if array_output.ndim == scalar_ndim:
            has_scalar_mode = False
        elif scalar_ndim != array_output.ndim - 1:
            raise ValueError("There is something wrong in the output dimensions of the function.")
    except ValueError:
        accept_arrays = False

    return accept_arrays, has_scalar_mode, is_constant


def shift_array(arr: np.ndarray, n: int, fill_value: float = 0):
    shifted = np.roll(arr, n, axis=-1)  # Shift the array
    if n > 0:
        shifted[..., :n] = fill_value  # Fill first n elements
    elif n < 0:
        shifted[..., n:] = fill_value  # Fill last n elements
    return shifted


def convert_to_string(content):
    try:
        if inspect.isfunction(content):
            name = getattr(content, "__name__")
            if name != "<lambda>":
                return name + str(inspect.signature(content))
            fun_string = inspect.getsourcelines(content)[0][0].strip()
            if "temporal(" in fun_string:
                lambda_content = fun_string.split("temporal(")[1].strip()[0:-1]
                return lambda_content
            for word in ["execute_on"]:
                if word in fun_string:
                    start_index = fun_string.find(word, len(word))
                    lambda_content = ", ".join(fun_string[start_index:].split(",")[1:])[:-1].strip()
                    return lambda_content
            if "=" in fun_string:
                fun_string = fun_string.split("=")[1].strip()
            return fun_string
        elif inspect.isclass(content):
            return content.__repr__()
    except Exception as e:
        warnings.warn(f"Developer error: Failed to create a string representation from {content} because of error {e}.")
    return str(content)


def add_necessary_brackets(expression: str) -> str:
    operators = ["+", "-", "=", "<", ">", "and", "or", "not", "xor"]
    begin = expression.split("(")[0]
    end = expression.split(")")[-1]
    if any(op in begin for op in operators) or any(op in end for op in operators):
        return f"({expression})"
    else:
        return expression


def is_custom_class(obj: Any) -> bool:
    # Check if the object is a built-in type like list, dict, scalar, or ndarray
    if isinstance(obj, (list, dict, int, float, np.ndarray, bool, str, types.FunctionType, types.LambdaType)):
        return False

    # Check if the object is an instance of a custom class
    if isinstance(obj, object):
        return True

    return False


def operator_call(obj, /, *args, **kwargs):
    """operator.call function source code copy in order to be used with Python version <3.11"""
    return obj(*args, **kwargs)



import enum
import functools
import inspect
import operator
import time
import warnings
from collections import abc
from copy import copy, deepcopy
from typing import overload, Literal, Iterable, Dict, Tuple, List, Any
from pathlib import Path
from typing import Callable, Union, TypeVar, Generic

import numpy as np
from numpy.typing import NDArray
from typing_extensions import ParamSpec
from cachebox import LRUCache

from .solver_utils import *
from .utils import add_necessary_brackets, convert_to_string, operator_call, shift_array, vectorize_source

if TYPE_CHECKING:
    import pandas as pd

T = TypeVar("T")
P = ParamSpec("P")

TriggerType = Union["CrossTriggerVar", "TemporalVar[bool]"]


class Solver:
    def __init__(self):
        self.dim = 0
        self.vars: List[TemporalVar] = []
        # All the scalars variables that are an input of an integrate function
        self.feed_vars: List[TemporalVar] = []
        # All the scalars variables that are an output of an integrate function
        self.integrated_vars: List[IntegratedVar] = []

        self.events: List[Event] = []
        self.cross_triggers: List[CrossTriggerVar] = []
        self.t_current: float = 0
        self.t = []
        self.y = None
        self.solved = False
        self.saved_vars: Dict[str, TemporalVar] = {}
        self.named_vars: Dict[str, TemporalVar] = {}
        self.vars_to_plot: Dict[str, TemporalVar] = {}
        self.status: Union[int, None] = None

        # Special solver TemporalVars
        self.time_variable: TemporalVar[float] = TemporalVar(self, lambda t: t, "Time")

    def integrate(self, input_value: "TemporalVar[T]", x0: T, minimum: Union[T, "TemporalVar[T]", None] = None,
                  maximum: Union[T, "TemporalVar[T]", None] = None) -> "IntegratedVar[T]":
        """
        Integrate the input value starting from the initial condition x0.

        :param input_value: The value to be integrated.
        :param x0: The initial condition for the integration.
        :return: The integrated TemporalVar.
        """
        if not isinstance(input_value, TemporalVar):
            input_value = TemporalVar(self, input_value)
        elif input_value.is_discrete:
            raise NotImplementedError("Discrete integration is not supported yet.")

        integrated_structure = self._get_integrated_structure(input_value, x0, minimum, maximum)
        integrated_variable = IntegratedVar(
            self,
            integrated_structure,
            f"#INTEGRATE {get_expression(input_value)}",
            x0
        )
        return integrated_variable

    def _get_integrated_structure(self, data, x0, minimum, maximum):
        if data.output_type is np.ndarray:
            if not isinstance(maximum, np.ndarray):
                maximum = np.full(data.shape, maximum)
            if not isinstance(minimum, np.ndarray):
                minimum = np.full(data.shape, minimum)
            result = np.empty(data.shape, dtype=object)
            for idx in np.ndindex(data.shape):
                result[idx] = self._get_integrated_structure(data[idx], np.array(x0)[idx], minimum[idx], maximum[idx])
            return result
        elif data.output_type is dict:
            if not isinstance(maximum, dict):
                maximum = {key: maximum for key in data.keys()}
            if not isinstance(minimum, dict):
                minimum = {key: minimum for key in data.keys()}

            return {
                key: self._get_integrated_structure(value, x0[key], minimum[key], maximum[key])
                for key, value in data.items()
            }

        return self._add_integration_variable(data, x0, minimum, maximum)

    def _add_integration_variable(self, var: Union["TemporalVar[T]", T], x0: T, minimum: Union["TemporalVar[T]", T],
                                  maximum: Union["TemporalVar[T]", T]) -> "IntegratedVar[T]":
        self.feed_vars.append(var)
        # Manage min and max
        if maximum is None:
            maximum = np.inf
        if minimum is None:
            minimum = -np.inf

        # Evaluate bounds at t = 0
        low_0 = minimum(0, self.x0) if isinstance(minimum, TemporalVar) else minimum
        up_0 = maximum(0, self.x0) if isinstance(maximum, TemporalVar) else maximum
        if not low_0 <= x0 <= up_0:
            raise ValueError(
                f"x0 = {x0} is outside the specified bounds [{low_0} ; {up_0}] at t=0. Please provide a value within these bounds.")

        # Add integration value
        integrated_variable = IntegratedVar(
            self,
            lambda t, y, idx=self.dim: y[idx] if y.ndim == 1 else y[..., idx],
            f"#INTEGRATE {get_expression(var)}",
            x0,
            minimum,
            maximum,
            self.dim,
        )
        self.integrated_vars.append(integrated_variable)
        self.dim += 1
        return integrated_variable

    def solve(
            self,
            t_end: float,
            method="RK45",
            time_step=None,
            t_eval=None,
            include_crossing_times: bool = True,
            plot: bool = True,
            verbose: bool = False,
            **options,
    ) -> None:
        """
        Solve the equations of the dynamical system through an integration scheme.

        :param t_end: Time at which the integration stops.
        :param method: Integration method to use. Default is 'RK45'.
        :param time_step: Time step for the integration. If None, use points selected by the solver.
        :param t_eval: Times at which to store the computed solution. If None, use points selected by the solver.
        :param plot: Plot the variables that called the "to_plot()" method.
        :param options: Additional options for the solver. See https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html.
        """
        self._get_remaining_named_variables()
        # Check if LoopNodes has been set
        for loop_node in (var for var in self.vars if isinstance(var, LoopNode)):
            if not loop_node.is_valid():
                raise Exception(
                    f"The value has not been set in loop node {loop_node.name}. Please call the 'loop_into()' method.\n"
                    "If this is intentional, instantiate the LoopNode object with parameter 'strict = False' to "
                    "disable this exception.")
        # Reinit values
        [var.clear() for var in self.vars]
        [e.clear() for e in self.events]
        start = time.time()
        # Set t_eval
        if time_step is not None and t_eval is None:
            t_eval = np.arange(0, t_end + time_step / 2,
                               time_step)  # Add half a time step to get an array that stops on t_end

        res = self._solve_ivp((0, t_end), self.x0, method=method, t_eval=t_eval,
                              include_crossing_times=include_crossing_times, **options)
        if not res.success:
            raise Exception(res.message)

        self.solved = True
        if verbose:
            output_str = f"Solving time = {time.time() - start} s\n"
            if self.events:
                output_str += f"Number of triggered events = {sum(len(t) for t in res.t_events)}\n"
            print(output_str)
        if plot:
            self.plot()

    @property
    def x0(self):
        return np.array([v.x0 for v in self.integrated_vars])

    def plot(self):
        """
        Plot the variables that have been marked for plotting.
        """
        if not self.vars_to_plot:
            return

        import matplotlib.pyplot as plt
        # Plot data
        plt.figure("Results for VIP")
        for variable_name, var in self.vars_to_plot.items():
            plt.plot(var.t, var.values, label=variable_name)
        # Label and axis
        plt.title("Simulation results")
        plt.xlabel("Time (s)")
        plt.legend()
        plt.xlim(0, self.t[-1])
        plt.grid()
        plt.tight_layout()
        plt.show()

    def explore(
            self,
            f: Callable,
            t_end: float,
            bounds=(),
            time_step: float = None,
            title: str = "",
    ):
        """
        Explore the function f over the given bounds and solve the system until t_end.
        This function needs the sliderplot package.

        :param title: Title of the plot
        :param time_step: Time step of the simulation
        :param f: The function to explore.
        :param t_end: Time at which the integration stops.
        :param bounds: Bounds for the exploration.
        """
        from sliderplot import sliderplot

        def wrapper(*args, **kwargs):
            self.clear()
            outputs = f(*args, **kwargs)
            self.solve(t_end, time_step=time_step)
            transformed_outputs = self._unwrap_leaves(outputs)
            return transformed_outputs

        functools.update_wrapper(wrapper, f)
        sliderplot(
            wrapper,
            bounds,
            page_title="vip-ivp",
            titles=[title],
            axes_labels=(("Time (s)", ""),),
        )

    def clear(self):
        """
        Clear stored information.
        """
        self.__init__()

    def _dy(self, t, y):
        result_list = []
        for var in self.feed_vars:
            try:
                result = var(t, y)
                result_list.append(result)
            except RecursionError:
                raise RecursionError(
                    f"An algebraic loop has been detected when trying to compute the value of variable {var.name}.\n"
                    f"Make sure that a variable does not reference itself in `.loop_into()` methods."
                )
        return result_list

    def _unwrap_leaves(self, outputs):
        """
        Transform all TemporalVar in an iterable into (x.t, x.values) pairs.

        :param outputs: The outputs to transform.
        :return: The transformed outputs.
        """
        if isinstance(outputs, TemporalVar):
            return outputs.t, outputs.values
        else:
            return list(map(self._unwrap_leaves, (el for el in outputs)))

    def _get_remaining_named_variables(self):
        frame = inspect.currentframe().f_back
        while (frame.f_locals.get("self")
               and (isinstance(frame.f_locals.get("self"), TemporalVar)
                    or isinstance(frame.f_locals.get("self"), Solver))
               or Path(frame.f_code.co_filename).as_posix().endswith("vip_ivp/api.py")):
            frame = frame.f_back
        local_variables = frame.f_locals
        for key, value in local_variables.items():
            if isinstance(value, TemporalVar) and key not in self.named_vars:
                value.name = key
                self.named_vars[key] = value

    CROSSING_TOLERANCE = 1e-13

    def _solve_ivp(
            self,
            t_span,
            y0,
            method="RK45",
            t_eval=None,
            dense_output=False,
            vectorized=False,
            include_crossing_times=True,
            **options,
    ):
        if method not in METHODS and not (
                inspect.isclass(method) and issubclass(method, OdeSolver)
        ):
            raise ValueError(f"`method` must be one of {METHODS} or OdeSolver class.")

        t0, tf = map(float, t_span)

        y0 = self._bound_sol(t0, y0)

        if t_eval is not None:
            t_eval = np.asarray(t_eval)
            if t_eval.ndim != 1:
                raise ValueError("`t_eval` must be 1-dimensional.")

            if np.any(t_eval < min(t0, tf)) or np.any(t_eval > max(t0, tf)):
                raise ValueError("Values in `t_eval` are not within `t_span`.")

            d = np.diff(t_eval)
            if tf > t0 and np.any(d <= 0) or tf < t0 and np.any(d >= 0):
                raise ValueError("Values in `t_eval` are not properly sorted.")

            if tf > t0:
                t_eval_i = 0
            else:
                # Make order of t_eval decreasing to use np.searchsorted.
                t_eval = t_eval[::-1]
                # This will be an upper bound for slices.
                t_eval_i = t_eval.shape[0]

        if method in METHODS:
            method = METHODS[method]

        if t_eval is None:
            self.t = [t0]
            self.y = [y0]
        elif t_eval is not None and dense_output:
            self.t = []
            ti = [t0]
            self.y = []
        else:
            self.t = []
            self.y = []

        solver = method(self._dy, t0, y0, tf, vectorized=vectorized, **options)

        t_events = []
        y_events = []
        [c.evaluate(t0, y0) for c in self.cross_triggers]

        interpolants = []

        self.status = None
        while self.status is None:
            message = solver.step()

            if solver.status == "failed":
                self.status = -1
                break

            t_old = solver.t_old
            t = solver.t
            y = self._bound_sol(t, solver.y)

            events = self.get_events(t)

            if dense_output:
                sol = lambda t: self._bound_sol(t, solver.dense_output()(t))
                interpolants.append(sol)
            else:
                sol = None

            if events or self.cross_triggers:
                if sol is None:
                    sol = self._sol_wrapper(solver.dense_output())

                # Create list of time sample where to check triggers and events
                t_eval_i_new = np.searchsorted(t_eval, t, side="right")
                t_eval_step = t_eval[:t_eval_i_new]
                t_eval_step = t_eval_step[t_eval_step > t_old]
                t_eval_step = list(np.unique([*t_eval_step, t]))

                # Prepare data for crossing detection
                tc_lower = t_old
                g = [c.previous_value for c in self.cross_triggers]
                directions = np.array([c.direction for c in self.cross_triggers])

                # Prevent events that triggered at the previous step to trigger again in this step, because its g_new is at 0 so
                # an irrelevant zero-crossing is sure to occur.
                previous_triggers_mask = np.array([not t_old in c.t_triggers for c in self.cross_triggers])

                discontinuity = False
                t_crossings = []

                while len(t_eval_step):
                    t_check = t_eval_step.pop(0)
                    y_check = sol(t_check)
                    # Detect crossing first
                    if self.cross_triggers:
                        g_new = [c.function(t_check, y_check) for c in self.cross_triggers]
                        active_crossing_indices = find_active_events_in_step(g, g_new, directions,
                                                                             previous_triggers_mask,
                                                                             self.CROSSING_TOLERANCE)
                        # If a crossing has been detected:
                        if active_crossing_indices.size > 0:
                            # Get the roots of each crossing
                            tc_upper = t_check
                            # Handle crossing by computing roots
                            active_crossings = [self.cross_triggers[idx] for idx in active_crossing_indices]
                            roots = [solve_event_equation(
                                c, sol, tc_lower, tc_upper, is_discrete(c),
                                self.CROSSING_TOLERANCE) for c in active_crossings]
                            roots = np.asarray(roots)
                            # Change the current t_check with the earliest trigger.
                            t_trigger = np.min(roots)
                            triggered_signals = [active_crossings[i] for i, root in enumerate(roots) if
                                                 root == t_trigger]
                            for signal in triggered_signals:
                                signal.t_triggers.append(t_trigger)
                                signal.previous_value = 0
                            g = [c.previous_value for c in self.cross_triggers]
                            previous_triggers_mask = np.array(
                                [not t_trigger in c.t_triggers for c in self.cross_triggers])
                            # Replace current time with trigger time and reevaluate the current t_check in the next loop
                            if t_trigger != t_check:
                                t_eval_step.insert(0, t_check)
                                t_check = t_trigger
                                y_check = sol(t_check)
                            tc_lower = t_trigger
                            t_crossings.append(t_trigger)
                        elif previous_triggers_mask is not None:
                            # Disable the preventing of zero-crossing from previously triggered events
                            g = g_new
                            tc_lower = t_check
                            previous_triggers_mask = None
                    first_iteration = False

                    # Detect events
                    if events:
                        y_before_event = np.asarray(deepcopy(y_check))
                        triggered_events = []
                        events_to_trigger = [e for e in self.events if e(t_check, y_check)]
                        for e in events_to_trigger:
                            e.execute_action(t_check, y_check)  # This can modify y_check value
                            triggered_events.append(e)
                        # Check for discontinuous action
                        discontinuity = not np.array_equal(y_before_event, np.asarray(y_check))
                        if discontinuity:
                            # Create a loop to check if other events has triggered because of modification of ye
                            while True:
                                events_to_trigger = [e for e in self.events if
                                                     e not in triggered_events and e(t_check, y_check)]
                                if events_to_trigger:
                                    for e in events_to_trigger:
                                        e.execute_action(t_check, y_check)  # This can modify y_check value
                                        triggered_events.append(e)
                                else:
                                    break
                            # End also crossing and event loop
                            break
                        # Handle termination
                        if self.status is not None:
                            break

                # Reset the solver and events evaluation to begin at te for the next step
                if discontinuity or self.status is not None:
                    t = t_check
                    y = y_check
                if discontinuity:
                    solver = method(self._dy, t, y, tf, vectorized=vectorized, **options)

            [c.evaluate(t, y) for c in self.cross_triggers]

            self.t_current = t
            if t_eval is None:
                self.t.append(t)
                self.y.append(y)
            else:
                # The value in t_eval equal to t will be included.
                if solver.direction > 0:
                    t_eval_i_new = np.searchsorted(t_eval, t, side="right")
                    t_eval_step = t_eval[t_eval_i:t_eval_i_new]
                else:
                    t_eval_i_new = np.searchsorted(t_eval, t, side="left")
                    # It has to be done with two slice operations, because
                    # you can't slice to 0th element inclusive using backward
                    # slicing.
                    t_eval_step = t_eval[t_eval_i_new:t_eval_i][::-1]

                # Include crossing times
                if include_crossing_times and self.cross_triggers and t_crossings:
                    t_eval_step = np.unique(np.concatenate((t_eval_step, t_crossings)))

                if t_eval_step.size > 0:
                    if sol is None:
                        sol = self._sol_wrapper(solver.dense_output())
                    self.t.extend(t_eval_step)
                    if self.dim != 0:
                        self.y.extend(sol(t_eval_step))
                    else:
                        self.y.extend([0] * len(t_eval_step))
                    t_eval_i = t_eval_i_new

                    if self.events and discontinuity:
                        if t_eval_step[-1] == t_check:
                            self.y[-1] = y_check

            if t_eval is not None and dense_output:
                ti.append(t)

            if solver.status == "finished":
                self.status = 0

        message = MESSAGES.get(self.status, message)
        if self.events:
            t_events = [np.asarray(e.t_events) for e in events]
            # y_events = [np.asarray(e.y_events) for e in events]

        if self.t:
            self.t = np.array(self.t)
            self.y = np.array(self.y)

        if dense_output:
            if t_eval is None:
                sol = OdeSolution(
                    self.t,
                    interpolants,
                    alt_segment=True if method in [BDF, LSODA] else False,
                )
            else:
                sol = OdeSolution(
                    ti,
                    interpolants,
                    alt_segment=True if method in [BDF, LSODA] else False,
                )
        else:
            sol = None

        return OdeResult(
            t=self.t,
            y=self.y,
            t_events=t_events,
            y_events=y_events,
            sol=sol,
            nfev=solver.nfev,
            njev=solver.njev,
            nlu=solver.nlu,
            status=self.status,
            message=message,
            success=self.status >= 0,
        )

    def _bound_sol(self, t, y: NDArray):
        upper, lower = self._get_bounds(t, y)
        y_bounded_max = np.where(y <= upper, y, upper)
        y_bounded = np.where(y_bounded_max >= lower, y_bounded_max, lower)
        return y_bounded

    def _get_bounds(self, t, y):
        upper = []
        lower = []
        for var in self.integrated_vars:
            maximum = var.maximum(t, y)
            minimum = var.minimum(t, y)
            if np.any(minimum > maximum):
                raise ValueError(
                    f"Lower bound {minimum} is greater than upper bound {maximum} a time {t} s "
                    f"for variable {var.name or var.expression}."
                )
            upper.append(maximum)
            lower.append(minimum)
        upper = np.moveaxis(np.array(upper), 0, -1)
        lower = np.moveaxis(np.array(lower), 0, -1)
        return upper, lower

    def _sol_wrapper(self, sol):
        def output_fun(t: Union[float, NDArray]):
            y = sol(t)
            if not np.isscalar(t):
                y = np.moveaxis(y, 0, -1)
            return self._bound_sol(t, y)

        return output_fun

    def get_events(self, t):
        event_list = [e for e in self.events]
        return event_list


class CallMode(enum.Enum):
    CALL_ARGS_FUN = 0
    CALL_FUN_RESULT = 1


class TemporalVar(Generic[T]):
    def __init__(
            self,
            solver: "Solver",
            source: Union[
                Callable[[Union[float, NDArray], NDArray], T],
                Callable[[Union[float, NDArray]], T],
                NDArray,
                Dict,
                float,
                Tuple
            ] = None,
            expression: str = None,
            child_cls=None,
            operator=None,
            call_mode: CallMode = CallMode.CALL_ARGS_FUN,
            is_discrete=False
    ):
        self.solver = solver
        self._output_type = None
        self._is_source = False
        self._call_mode = call_mode
        self.is_discrete = is_discrete
        # Recursive building
        self.operator = operator
        child_cls = child_cls or type(self)
        if self.operator is not None:
            self.source = source
        else:
            self._is_source = True
            if callable(source) and not isinstance(source, child_cls):
                n_args = len(inspect.signature(source).parameters)
                if n_args == 1:
                    self.source = lambda t, y: vectorize_source(source)(t)
                else:
                    self.source = lambda t, y: source(t, y)
            elif np.isscalar(source):
                self.source = source
                self._output_type = type(source)
            elif isinstance(source, (list, np.ndarray)):
                self._output_type = np.ndarray
                self.source = np.vectorize(lambda f: child_cls(solver, f))(
                    np.array(source)
                )
            elif isinstance(source, TemporalVar):
                vars(self).update(vars(source))
            elif isinstance(source, dict):
                self._output_type = dict
                self.source = {key: child_cls(solver, val) for key, val in source.items()}
            elif source is None:
                self.source = None
            else:
                raise ValueError(f"Unsupported type: {type(source)}.")

        self._values = None
        # Variable definition
        self._expression = convert_to_string(source) if expression is None else expression
        self.name = None

        # self.events: List[Event] = []

        self._cache = LRUCache(maxsize=4)

        self.solver.vars.append(self)

    def __call__(self, t: Union[float, NDArray], y: NDArray) -> T:
        # Handle dict in a recursive way
        if isinstance(self.source, dict):
            return {key: val(t, y) for key, val in self.source.items()}
        else:
            # Handle the termination leaves of the recursion
            if isinstance(t, np.ndarray):
                t_cache = tuple(t)
            else:
                t_cache = t
            if t_cache in self._cache:
                return self._cache[t_cache]
            elif isinstance(self.source, np.ndarray):
                if np.isscalar(t):
                    output = np.stack(np.frompyfunc(lambda f: f(t, y), 1, 1)(self.source))
                else:
                    output = np.stack(
                        np.frompyfunc(lambda f: f(t, y), 1, 1)(self.source.ravel())
                    ).reshape((*self.source.shape, *np.array(t).shape))
            elif self.operator is not None:
                if self.operator is operator_call and not np.isscalar(t):
                    output = np.array([self._resolve_operator(t[i], y[i]) for i in range(len(t))])
                    if output.ndim > 1:
                        output = np.moveaxis(output, 0, -1)
                else:
                    output = self._resolve_operator(t, y)
            else:
                if callable(self.source):
                    output = self.source(t, y)
                elif np.isscalar(t):
                    output = self.source
                else:
                    output = np.full(len(t), self.source)
            if self.solver.solved:
                self._cache[t_cache] = output
        return output

    def _resolve_operator(self, t, y):
        if self._call_mode == CallMode.CALL_ARGS_FUN:
            args = [x(t, y) if isinstance(x, TemporalVar) else x for x in self.source if
                    not isinstance(x, dict)]
            kwargs = {k: v for d in [x for x in self.source if isinstance(x, dict)] for k, v in d.items()}
            kwargs = {k: (x(t, y) if isinstance(x, TemporalVar) else x) for k, x in kwargs.items()}
            return self.operator(*args, **kwargs)
        elif self._call_mode == CallMode.CALL_FUN_RESULT:
            args = [x for x in self.source if not isinstance(x, dict)]
            kwargs = {k: v for d in [x for x in self.source if isinstance(x, dict)] for k, v in d.items()}
            return self.operator(*args, **kwargs)(t, y)
        else:
            raise ValueError(f"Unknown call mode: {self._call_mode}.")

    @property
    def values(self) -> NDArray:
        if not self.solver.solved:
            raise Exception(
                "The differential system has not been solved. "
                "Call the solve() method before inquiring the variable values."
            )
        try:
            if self._values is None:
                self._values = self(self.t, self.solver.y)
        except RecursionError:
            raise RecursionError(
                f"An algebraic loop has been detected when trying to compute the value of variable {self.name}.\n"
                f"Make sure that a variable does not reference itself in `.loop_into()` methods."
            )
        return self._values

    @property
    def t(self) -> NDArray:
        if not self.solver.solved:
            raise Exception(
                "The differential system has not been solved. "
                "Call the solve() method before inquiring the time variable."
            )
        return np.asarray(self.solver.t)

    @property
    def output_type(self):
        if self._output_type is None:
            self._output_type = type(self._first_value())
        return self._output_type

    def save(self, name: str) -> None:
        """
        Save the temporal variable with a name.

        :param name: Key to retrieve the variable.
        """
        if name in self.solver.saved_vars:
            warnings.warn(
                f"A variable with name {name} already exists. Its value has been overridden."
            )
        self.solver.saved_vars[name] = self

    def to_plot(self, name: str = None) -> None:
        """
        Add the variable to the plotted data on solve.

        :param name: Name of the variable in the legend of the plot.
        """
        if name is None:
            if self.name is None:
                get_expression(self)
            name = self.name
        if self.output_type is np.ndarray:
            [
                self[idx].to_plot(f"{name}[{', '.join(str(i) for i in idx)}]")
                for idx in np.ndindex(self.shape)
            ]
            return
        elif isinstance(self.source, dict):
            [self[key].to_plot(f"{name}[{key}]") for key in self._first_value().keys()]
            return
        self.solver.vars_to_plot[name] = self

    @classmethod
    def from_scenario(
            cls,
            solver: "Solver",
            scenario_table: "pd.DataFrame",
            time_key: str,
            interpolation_kind="linear",
    ) -> "TemporalVar":
        from scipy.interpolate import interp1d

        variables = {}
        for col in scenario_table.columns:
            if col == time_key:
                continue
            fun = interp1d(
                scenario_table[time_key],
                scenario_table[col],
                kind=interpolation_kind,
                bounds_error=False,
                fill_value=(scenario_table[col].iat[0], scenario_table[col].iat[-1]),
            )
            variables[col] = fun
        return cls(solver, variables)

    def delayed(self, delay: int, initial_value: T = 0) -> "TemporalVar[T]":
        """
        Create a delayed version of the TemporalVar.
        :param delay: Number of solver steps by which the new TemporalVar is delayed.
        :param initial_value: Value of the delayed variable at the beginning when there is not any value for the original value.
        :return: Delayed version of the TemporalVar
        """
        if delay < 1:
            raise Exception("Delay accept only a positive step.")

        def create_delay(input_variable):
            def previous_value(t, y):
                if np.isscalar(t):
                    if len(input_variable.solver.t) >= delay:
                        index = np.searchsorted(input_variable.solver.t, t, "left")
                        # index = next((i for i, ts in enumerate(input_variable.solver.t) if t <= ts),
                        #              len(input_variable.solver.t))
                        if index - delay < 0:
                            return initial_value
                        previous_t = input_variable.solver.t[index - delay]
                        previous_y = input_variable.solver.y[index - delay]

                        return input_variable(previous_t, previous_y)
                    else:
                        return initial_value
                else:
                    delayed_t = shift_array(t, delay, 0)
                    delayed_y = shift_array(y, delay, initial_value)
                    return input_variable(delayed_t, delayed_y)

            return previous_value

        if 2 * delay > self._cache.maxsize:
            self._cache = LRUCache(maxsize=2 * delay)

        return TemporalVar(self.solver, (create_delay, self),
                           expression=f"#DELAY({delay}) {get_expression(self)}",
                           operator=operator_call,
                           call_mode=CallMode.CALL_FUN_RESULT,
                           is_discrete=True)

    def derivative(self, initial_value=0) -> "TemporalVar[T]":
        """
        Return the derivative of the Temporal Variable.

        Warning: Contrary to integration, this derivative method does not guarantee precision. Use it only as an escape
        hatch.
        :param initial_value: value at t=0
        :return: TemporalVar containing the derivative.
        """
        # Warn the user not to abuse the differentiate function
        warnings.warn("It is recommended to use 'integrate' instead of 'differentiate' for solving IVPs, "
                      "because the solver cannot guarantee precision when computing derivatives.\n"
                      "If you choose to use 'differentiate', consider using a smaller step size for better accuracy.",
                      category=UserWarning, stacklevel=2)

        previous = self.delayed(1)
        time_value = self.solver.time_variable
        previous_time = time_value.delayed(1)
        d_y = self - previous
        d_t = time_value - previous_time
        derived_value = temporal_var_where(self.solver, time_value != 0, np.divide(d_y, d_t, where=d_t != 0),
                                           initial_value)
        derived_value._expression = f"#D/DT {get_expression(self)}"
        return derived_value

    def m(self, method: Callable[P, T]) -> Callable[P, "TemporalVar[T]"]:
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> TemporalVar:
            inputs_expr = [get_expression(inp) if isinstance(inp, TemporalVar) else str(inp) for inp in args]
            kwargs_expr = [
                f"{key}={get_expression(value) if isinstance(value, TemporalVar) else str(value)}"
                for key, value in kwargs.items()
            ]
            expression = f"{method.__name__}({', '.join(inputs_expr)}"
            if kwargs_expr:
                expression += ", ".join(kwargs_expr)
            expression += ")"
            return TemporalVar(self.solver, (method, self, *args, kwargs),
                               expression=expression, operator=operator_call)

        functools.update_wrapper(wrapper, method)
        return wrapper

    def crosses(self, value: Union["TemporalVar[T]", T],
                direction: Literal["rising", "falling", "both"] = "both") -> "CrossTriggerVar":
        """
        Create a signal that triggers when the specified crossing occurs.

        :param value: Value to be crossed to cause the triggering.
        :param direction: Direction of the crossing.
        :return: TriggerVar
        """
        if self.output_type in (bool, np.bool, str):
            crossed_variable = self == value
        elif issubclass(self.output_type, abc.Iterable):
            raise ValueError(
                "Can not apply crossing detection to a variable containing a collection of values because it is ambiguous."
            )
        else:
            crossed_variable = self - value
        expression = f"#CROSSING_BETWEEN {self.name} AND {value.name if isinstance(value, TemporalVar) else value}"
        trigger_var = CrossTriggerVar(self.solver, crossed_variable, direction, expression)
        return trigger_var

    def clear(self):
        self._values = None
        self._cache.clear()

    def _first_value(self):
        return self(0, self.solver.x0 if len(self.solver.x0) else 0)

    def _from_arg(self, value: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        """
        Return a TemporalVar from an argument value. If the argument is already a TemporalVar, return it. If not, create a TemporalVar from the value.
        """
        if isinstance(value, TemporalVar) and value.solver is self.solver:
            return value
        return TemporalVar(self.solver, value)

    def __copy__(self):
        return TemporalVar(self.solver, self.source, self.expression, operator=self.operator)

    def __add__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{get_expression(self)} + {get_expression(other)}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.add
        )

    def __radd__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{get_expression(other)} + {get_expression(self)}"
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.add
        )

    def __sub__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = (
            f"{get_expression(self)} - {add_necessary_brackets(get_expression(other))}"
        )
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.sub
        )

    def __rsub__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = (
            f"{get_expression(other)} - {add_necessary_brackets(get_expression(self))}"
        )
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.sub
        )

    def __mul__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(self))} * {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.mul
        )

    def __rmul__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(other))} * {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.mul
        )

    def __truediv__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(self))} / {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.truediv
        )

    def __rtruediv__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(other))} / {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.truediv
        )

    def __floordiv__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(self))} // {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.floordiv
        )

    def __rfloordiv__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(other))} // {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.floordiv
        )

    def __mod__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(self))} % {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.mod
        )

    def __rmod__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(other))} % {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.mod
        )

    def __pow__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(self))} ** {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.pow
        )

    def __rpow__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[T]":
        expression = f"{add_necessary_brackets(get_expression(other))} ** {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._from_arg(other), self),
            expression=expression,
            operator=operator.pow
        )

    def __pos__(self) -> "TemporalVar[T]":
        expression = f"+ {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self,),
            expression=expression,
            operator=operator.pos
        )

    def __neg__(self) -> "TemporalVar[T]":
        expression = f"- {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self,),
            expression=expression,
            operator=operator.neg
        )

    def __abs__(self) -> "TemporalVar[T]":
        expression = f"abs({get_expression(self)})"
        return TemporalVar(
            self.solver,
            (self,),
            expression=expression,
            operator=operator.abs
        )

    @overload
    def __eq__(self, other: "TemporalVar[NDArray[T]]") -> "TemporalVar[NDArray[bool]]":
        ...

    @overload
    def __eq__(self, other: Any) -> "TemporalVar[bool]":
        ...

    def __eq__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} == {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.eq
        )

    @overload
    def __ne__(self, other: "TemporalVar[NDArray[T]]") -> "TemporalVar[NDArray[bool]]":
        ...

    @overload
    def __ne__(self, other: Any) -> "TemporalVar[bool]":
        ...

    def __ne__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} != {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.ne
        )

    @overload
    def __lt__(self, other: "TemporalVar[NDArray[T]]") -> "TemporalVar[NDArray[bool]]":
        ...

    @overload
    def __lt__(self, other: Any) -> "TemporalVar[bool]":
        ...

    def __lt__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} < {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.lt
        )

    @overload
    def __le__(self, other: "TemporalVar[NDArray[T]]") -> "TemporalVar[NDArray[bool]]":
        ...

    @overload
    def __le__(self, other: Any) -> "TemporalVar[bool]":
        ...

    def __le__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} <= {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.le
        )

    @overload
    def __gt__(self, other: "TemporalVar[NDArray[T]]") -> "TemporalVar[NDArray[bool]]":
        ...

    @overload
    def __gt__(self, other: Any) -> "TemporalVar[bool]":
        ...

    def __gt__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} > {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.gt
        )

    @overload
    def __ge__(self, other: "TemporalVar[NDArray[T]]") -> "TemporalVar[NDArray[bool]]":
        ...

    @overload
    def __ge__(self, other: Any) -> "TemporalVar[bool]":
        ...

    def __ge__(self, other: Union["TemporalVar[T]", T]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} >= {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self, self._from_arg(other)),
            expression=expression,
            operator=operator.ge
        )

    @staticmethod
    def _apply_logical(logical_fun: Callable, a, b):
        result = logical_fun(a, b)
        if result.size == 1:
            result = result.item()
        return result

    def __and__(self, other: Union[bool, "TemporalVar[bool]"]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} and {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self._apply_logical, np.logical_and, self, self._from_arg(other)),
            expression,
            operator=operator_call
        )

    def __rand__(self, other: Union[bool, "TemporalVar[bool]"]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(other))} and {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._apply_logical, np.logical_and, self._from_arg(other), self),
            expression,
            operator=operator_call
        )

    def __or__(self, other: Union[bool, "TemporalVar[bool]"]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} or {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self._apply_logical, np.logical_or, self, self._from_arg(other)),
            expression,
            operator=operator_call
        )

    def __ror__(self, other: Union[bool, "TemporalVar[bool]"]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(other))} or {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._apply_logical, np.logical_or, self._from_arg(other), self),
            expression,
            operator=operator_call
        )

    def __xor__(self, other: Union[bool, "TemporalVar[bool]"]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(self))} xor {add_necessary_brackets(get_expression(other))}"
        return TemporalVar(
            self.solver,
            (self._apply_logical, np.logical_xor, self, self._from_arg(other)),
            expression,
            operator=operator_call
        )

    def __rxor__(self, other: Union[bool, "TemporalVar[bool]"]) -> "TemporalVar[bool]":
        expression = f"{add_necessary_brackets(get_expression(other))} xor {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._apply_logical, np.logical_xor, self._from_arg(other), self),
            expression,
            operator=operator_call
        )

    @staticmethod
    def _logical_not(a):
        result = np.logical_not(a)
        if result.size == 1:
            result = result.item()
        return result

    def __invert__(self) -> "TemporalVar[bool]":
        expression = f"not {add_necessary_brackets(get_expression(self))}"
        return TemporalVar(
            self.solver,
            (self._logical_not, self),
            expression,
            operator=operator_call
        )

    def __getitem__(self, item):
        expression = f"{add_necessary_brackets(get_expression(self))}[{item}]"
        return TemporalVar(
            self.solver,
            (self, item),
            expression=expression,
            operator=operator.getitem
        )

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs) -> "TemporalVar":
        inputs_expr = [
            get_expression(inp) if isinstance(inp, TemporalVar) else str(inp)
            for inp in inputs
        ]
        kwargs_expr = [
            f"{key}={get_expression(value) if isinstance(value, TemporalVar) else str(value)}"
            for key, value in kwargs.items()
        ]
        expression = f"{ufunc.__name__}({', '.join(inputs_expr)}"
        if kwargs:
            expression += f", {', '.join(kwargs_expr)}"
        expression += ")"
        if method == "__call__":
            return TemporalVar(
                self.solver,
                (ufunc, *inputs, kwargs),
                expression=expression,
                operator=operator_call
            )

        return NotImplemented

    def __bool__(self):
        raise ValueError("The truth value of a Temporal Variable is ambiguous. Use vip.where() instead.")

    @property
    def expression(self):
        return self._expression

    def __repr__(self) -> str:
        if self.solver.solved:
            output = f"{self.name or self._expression}"
            if self._values is not None:
                output += f" = {self._values}"
            return output
        else:
            return f"{self._expression}"

    # NumPy arrays utility methods
    @property
    def shape(self):
        result = self._first_value()
        if isinstance(result, np.ndarray):
            return result.shape
        raise AttributeError("hhape attribute does not exist because this variable does not contain a NumPy array.")

    # Dict utility methods
    def keys(self):
        result = self._first_value()
        if isinstance(result, dict):
            return result.keys()
        raise AttributeError("keys() method does not exist because this variable does not contain a dict.")

    def items(self):
        result = self._first_value()
        if isinstance(result, dict):
            key_list = list(result.keys())
            value_list = [self[key] for key in key_list]
            return zip(key_list, value_list)
        raise AttributeError("items() method does not exist because this variable does not contain a dict.")


def convert_args_to_temporal_var(solver: Solver, arg_list: Iterable) -> List[TemporalVar]:
    def convert(arg):
        if not isinstance(arg, TemporalVar):
            arg = TemporalVar(solver, arg)
        return arg

    return [convert(a) for a in arg_list]


def temporal_var_where(solver, condition: TemporalVar[bool], a: Union[T, TemporalVar[T]],
                       b: Union[T, TemporalVar[T]]) -> \
        TemporalVar[T]:
    def where(condition, a, b):
        result = np.where(condition, a, b)
        if result.size == 1:
            result = result.item()
        return result

    return TemporalVar(
        solver,
        (where, condition, a, b),
        expression=f"({get_expression(a)} if {get_expression(condition)} else {get_expression(b)})",
        operator=operator_call
    )


class LoopNode(TemporalVar[T]):
    def __init__(self, solver: "Solver", shape: Union[int, tuple[int, ...]] = None, strict: bool = True):
        if shape is not None:
            initial_value = np.zeros(shape)
        else:
            initial_value = 0
        self._input_var_content = None
        super().__init__(solver, 0, expression="", child_cls=TemporalVar)
        self._input_var: TemporalVar = TemporalVar(solver, initial_value)
        self._is_set = False
        self._is_strict = strict

    def loop_into(self, value: Union[TemporalVar[T], T, List], force: bool = False):
        """
        Set the input value of the loop node.

        :param force: Add the value to the loop node even if it has already been set.
        :param value: The value to add, can be a TemporalVar or a number.
        """
        if self._is_set and not force:
            raise Exception(
                "This Loop Node has already been set. If you want to add another value, use argument 'force = True'."
            )
        if not isinstance(value, TemporalVar):
            value = TemporalVar(self.solver, value)
        if not self._is_set:
            self._input_var = value
        else:
            self._input_var += value
        self._is_set = True
        self._expression = get_expression(self._input_var)
        self.source = self._input_var.source
        self.operator = self._input_var.operator

    @property
    def _input_var(self):
        return self._input_var_content

    @_input_var.setter
    def _input_var(self, value: TemporalVar[T]):
        self._input_var_content = value
        # Overwrite attributes except cache size
        cache = self._cache
        vars(self).update(vars(self._input_var_content))
        self._cache = cache

    def is_valid(self) -> bool:
        """
        Check if the Loop Node is ready to be solved.
        If the Loop Node uses strict mode, its value must be set.
        :return: True if valid, False if incorrect
        """
        return not self._is_strict or self._is_set


class IntegratedVar(TemporalVar[T]):
    def __init__(
            self,
            solver: "Solver",
            fun: Union[
                Callable[[Union[float, NDArray], NDArray], T], NDArray, dict
            ] = None,
            expression: str = None,
            x0: T = None,
            minimum: Union[TemporalVar[T], T] = -np.inf,
            maximum: Union[TemporalVar[T], T] = np.inf,
            y_idx: int = None
    ):
        self.x0 = x0
        self.maximum = convert_args_to_temporal_var(solver, (maximum,))[0]
        self.minimum = convert_args_to_temporal_var(solver, (minimum,))[0]

        self._y_idx = y_idx
        if isinstance(fun, IntegratedVar):
            self._y_idx = IntegratedVar.y_idx
        super().__init__(solver, fun, expression)

    def __getitem__(self, item) -> "IntegratedVar":
        return self.source[item]

    @property
    def y_idx(self):
        if isinstance(self.source, np.ndarray):
            return np.vectorize(lambda v: v.y_idx)(self.source)
        elif isinstance(self.source, dict):
            return {key: value.y_idx for key, value in self.source.items()}
        elif self._y_idx is not None:
            return self._y_idx
        raise ValueError("The argument 'y_idx' should be set for IntegratedVar containing a single value.")

    def reset_on(self, trigger: TriggerType, new_value: Union[TemporalVar[T], T]) -> "Event":
        """
        Create an action that, when its event is triggered, reset the IntegratedVar output to the specified value.
        :param trigger: Variable that triggers the reset
        :param new_value: Value at which the integrator output is reset to
        :return: Event.
        """
        if not isinstance(new_value, TemporalVar):
            new_value = TemporalVar(self.solver, new_value)

        def action_fun(t, y):
            def set_y0(idx, subvalue):
                if isinstance(idx, np.ndarray):
                    for arr_idx in np.ndindex(idx.shape):
                        y_idx = idx[arr_idx]
                        set_y0(y_idx, new_value[y_idx])
                elif isinstance(idx, dict):
                    for key, idx in idx.items():
                        y[idx] = new_value[key]
                        set_y0(idx, new_value[key])
                else:
                    y[idx] = subvalue(t, y)

            set_y0(self.y_idx, new_value)

        action = Action(action_fun, f"Reset {self.name} to {new_value.expression}")
        event = Event(self.solver, trigger, action)
        return event


class CrossTriggerVar(TemporalVar[bool]):
    _DIRECTION_MAP = {"rising": 1, "falling": -1, "both": 0}

    def __init__(self, solver: Solver, fun: TemporalVar, direction: Literal["rising", "falling", "both"] = "both",
                 expression: str = None):
        super().__init__(solver)
        self.direction = self._DIRECTION_MAP[direction]
        self.function = fun
        self._expression = expression

        self.t_triggers = []

        self.previous_value = None
        # Add to solver
        self.solver.cross_triggers.append(self)

    def __call__(self, t, y):
        if np.isscalar(t):
            return t in self.t_triggers
        else:
            return [self(t[i], y[i]) for i in range(len(t))]

    def evaluate(self, t, y):
        self.previous_value = self.function(t, y)


def get_expression(value) -> str:
    if isinstance(value, TemporalVar):
        frame = inspect.currentframe().f_back.f_back
        while (
                "self" in frame.f_locals
                and (
                        isinstance(frame.f_locals["self"], TemporalVar)
                        or isinstance(frame.f_locals["self"], Solver)
                )
                or Path(frame.f_code.co_filename).as_posix().endswith("vip_ivp/api.py")
        ):
            frame = frame.f_back
        found_key = next(
            (key for key, dict_value in frame.f_locals.items() if dict_value is value),
            None,
        )
        if found_key is not None:
            value.name = found_key
            value.solver.named_vars[found_key] = value
            return value.name
        return value.expression
    else:
        return str(value)


class Event:
    def __init__(self, solver: Solver, fun: TemporalVar[bool], action: Union["Action", Callable, None] = None):
        self.solver = solver
        self.function: TemporalVar = convert_args_to_temporal_var(self.solver, [fun])[0]
        self.action = convert_args_to_action([action])[0] if action is not None else None

        self.t_events = []

        self.solver.events.append(self)

    def __call__(self, t, y) -> bool:
        return bool(self.function(t, y))

    def __repr__(self):
        return (f"Event(On {repr(self.function)}, "
                f"{self.action or 'No action'})")

    def execute_action(self, t, y):
        if self.action is not None:
            self.action(t, y)
        self.t_events.append(t)

    def clear(self):
        self.t_events = []

    @property
    def count(self):
        return len(self.t_events)


class Action:
    def __init__(self, fun: Callable, expression: str = None):
        if isinstance(fun, TemporalVar):
            raise ValueError(
                "An action can not be a TemporalVar, because an action is a function with side effects, "
                "while a TemporalVar is a pure function."
            )
        if callable(fun):
            n_args = len(inspect.signature(fun).parameters)
            if n_args == 0:
                self.function = lambda t, y: fun()
            elif n_args == 1:
                self.function = lambda t, y: fun(t)
            else:
                self.function = lambda t, y: fun(t, y)
        self.expression = expression or convert_to_string(fun)

    def __call__(self, t, y):
        return self.function(t, y)

    def __add__(self, other: Union[Callable, "Action"]) -> "Action":
        if not isinstance(other, Action):
            other = Action(other)

        def added_fun(t, y):
            self(t, y)
            other(t, y)

        return Action(added_fun, f"{self.expression} + {other.expression}")

    def __repr__(self):
        return f"{self.expression}"


def convert_args_to_action(arg_list: Iterable) -> List[Action]:
    def convert(arg):
        if not isinstance(arg, Action):
            arg = Action(arg)
        return arg

    return [convert(a) for a in arg_list]



import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas as pd

from .base import *
from .utils import *

warnings.simplefilter("once")

_solver_list = []

T = TypeVar('T')
K = TypeVar("K")


@overload
def temporal(value: List[T]) -> TemporalVar[NDArray[T]]: ...


@overload
def temporal(value: Callable[[Union[float, NDArray]], T]) -> TemporalVar: ...


@overload
def temporal(value: int) -> TemporalVar[float]: ...


@overload
def temporal(value: T) -> TemporalVar[T]: ...


def temporal(value: Union[Callable[[NDArray], T], T]) -> TemporalVar[T]:
    """
    Create a Temporal Variable from a temporal function, a scalar value, a dict, a list or a NumPy array.

    If the input value is a list, the variable content will be converted to a NumPy array. As a consequence, a nested
    list must represent a valid rectangular matrix.

    :param value: A function f(t), a scalar value, a dict, a list or a NumPy array.
    :return: The created TemporalVar.
    """
    solver = _get_current_solver()
    return TemporalVar(solver, value)


def create_scenario(scenario_table: Union["pd.DataFrame", str, dict], time_key: str, interpolation_kind: str = "linear",
                    sep: str = ',') -> TemporalVar:
    """
    Creates a scenario from a given input table, which can be in various formats such as CSV, JSON, dictionary, or DataFrame.

    The maps in the scenario table are interpolated over time and converted into TemporalVar objects.
    The function processes the data and returns a TemporalVar containing a dictionary of TemporalVar objects.

    :param scenario_table: The input data, which can be one of the following formats:
        - A CSV file path (string)
        - A JSON file path (string)
        - A dictionary of data
        - A pandas DataFrame
    :param time_key: The key (column) to use as time for the scenario.
    :param interpolation_kind: Specifies the kind of interpolation as a string or as an integer specifying the order of
        the spline interpolator to use. The string has to be one of ‘linear’, ‘nearest’, ‘nearest-up’, ‘zero’, ‘slinear’,
        ‘quadratic’, ‘cubic’, ‘previous’, or ‘next’. ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer to a spline
        interpolation of zeroth, first, second or third order; ‘previous’ and ‘next’ simply return the previous or next
        value of the point; ‘nearest-up’ and ‘nearest’ differ when interpolating half-integers (e.g. 0.5, 1.5) in that
        ‘nearest-up’ rounds up and ‘nearest’ rounds down. Default is ‘linear’.
    :param sep: The separator to use when reading CSV files. Default is a comma.
    :return: A dictionary of TemporalVar objects representing the scenario, where the keys are the variables and the values are the corresponding TemporalVar instances.

    """
    import pandas as pd

    solver = _get_current_solver()
    if isinstance(scenario_table, str):
        if scenario_table.endswith(".csv"):
            input_data = pd.read_csv(scenario_table, sep=sep)
            print(input_data)
            return TemporalVar.from_scenario(solver, input_data, time_key, interpolation_kind)
        elif scenario_table.endswith(".json"):
            with open(scenario_table, "r") as file:
                dict_data = json.load(file)
            input_data = pd.DataFrame(dict_data)
            return TemporalVar.from_scenario(solver, input_data, time_key, interpolation_kind)
        else:
            raise ValueError("Unsupported file type")
    elif isinstance(scenario_table, dict):
        input_data = pd.DataFrame(scenario_table)
        return TemporalVar.from_scenario(solver, input_data, time_key, interpolation_kind)
    elif isinstance(scenario_table, pd.DataFrame):
        return TemporalVar.from_scenario(solver, scenario_table, time_key, interpolation_kind)
    else:
        raise ValueError("Unsupported input type")


@overload
def integrate(input_value: Union[int, float], x0: Union[int, float], minimum: Union[TemporalVar[T], T, None] = None,
              maximum: Union[TemporalVar[T], T, None] = None) -> IntegratedVar[float]: ...


@overload
def integrate(input_value: Union[TemporalVar[T], T], x0: Union[T, List], minimum: Union[TemporalVar[T], T, None] = None,
              maximum: Union[TemporalVar[T], T, None] = None) -> IntegratedVar[T]: ...


def integrate(input_value: Union[TemporalVar[T], T], x0: Union[T, List], minimum: Union[TemporalVar[T], T, None] = None,
              maximum: Union[TemporalVar[T], T, None] = None) -> IntegratedVar[T]:
    """
    Integrate the input value starting from the initial condition x0.

    The integrated output can be bounded with the **minimum** and **maximum** arguments.

    :param input_value: The value to be integrated, can be a TemporalVar or a number.
    :param x0: The initial condition for the integration.
    :param minimum: Lower integration bound. Can be a TemporalVar
    :param maximum: Higher integration bound. Can be a TemporalVar
    :return: The integrated TemporalVar.
    """
    solver = _get_current_solver()
    _check_solver_discrepancy(input_value, solver)
    integral_value = solver.integrate(input_value, x0, minimum, maximum)
    return integral_value


@overload
def loop_node(shape: None = None, strict: bool = True) -> LoopNode[float]: ...


@overload
def loop_node(shape: Union[int, tuple[int, ...]] = None, strict: bool = True) -> LoopNode[NDArray]: ...


def loop_node(shape: Union[int, tuple[int, ...]] = None, strict: bool = True) -> LoopNode:
    """
    Create a Loop Node. A Loop Node can accept new inputs through its "loop_into()" method after being instantiated.

    :param shape: Shape of the NumPy array contained in the Loop Node. If None, the Loop Node will contain a scalar.
    :param strict: Flag that triggers an error when the Loop Node has not been set before the solving.

    :return: The created LoopNode.
    """
    solver = _get_current_solver()
    return LoopNode(solver, shape, strict)


@overload
def where(condition: Union[TemporalVar[bool], bool], a: Union[TemporalVar, T], b: Union[TemporalVar, T]) -> TemporalVar[
    T]: ...


def where(condition: Union[TemporalVar[bool], bool], a: Union[TemporalVar, T],
          b: Union[TemporalVar, T]) -> TemporalVar[T]:
    """
    Create a conditional TemporalVar or a conditional Action.
    If condition is `True` at time $t$, the output value will have value **a**, else **b**.

    If **a** and **b** are TemporalVars or scalars, the output will be a TemporalVar.

    If **a** and **b** are Actions, the output will be an Action.

    :param condition: Condition to evaluate through time.
    :param a: Output value or Action to execute if the condition is `True` at time $t$
    :param b: Output value or Action to execute if the condition is `False` at time $t$
    :return: Conditional TemporalVar or conditional Action.
    """
    solver = _get_current_solver()
    return temporal_var_where(solver, condition, a, b)


P = ParamSpec("P")


def f(func: Callable[P, T]) -> Callable[P, TemporalVar[T]]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> TemporalVar:
        inputs_expr = [get_expression(inp) if isinstance(inp, TemporalVar) else str(inp) for inp in args]
        kwargs_expr = [
            f"{key}={get_expression(value) if isinstance(value, TemporalVar) else str(value)}"
            for key, value in kwargs.items()
        ]
        if inspect.isclass(func):
            expression = f"{func.__class__.__name__}({', '.join(inputs_expr)}"
        elif inspect.isfunction(func):
            expression = f"{func.__name__}({', '.join(inputs_expr)}"
        else:
            expression = f"{repr(func)}({', '.join(inputs_expr)}"
        if kwargs_expr:
            expression += ", ".join(kwargs_expr)
        expression += ")"

        return TemporalVar(_get_current_solver(), (func, *args, kwargs),
                           expression=expression, operator=operator_call)

    functools.update_wrapper(wrapper, func)
    return wrapper


def get_time_variable() -> TemporalVar[float]:
    """
    Return the time variable of the system.
    :return: Time variable that returns the time value of each simulation step.
    """
    solver = _get_current_solver()
    return solver.time_variable


# Events

def terminate_on(trigger: TriggerType) -> Event:
    solver = _get_current_solver()

    def action_terminate():
        solver.status = 1

    action = Action(action_terminate, "Terminate simulation")
    event = Event(solver, trigger, action)
    return event


def timeout_trigger(delay: float) -> CrossTriggerVar:
    solver = _get_current_solver()
    current_time = solver.t_current
    time_variable = get_time_variable()
    trigger = time_variable.crosses(current_time + delay)
    return trigger


def interval_trigger(delay: float) -> CrossTriggerVar:
    solver = _get_current_solver()
    current_time = solver.t_current
    time_variable = copy(get_time_variable())
    # Convert to a sine wave of period delay
    periodic_sine = where(time_variable < current_time, 0, np.sin(np.pi * (time_variable - current_time) / delay))
    trigger = periodic_sine.crosses(0)
    return trigger


def execute_on(trigger: TriggerType, f: Callable) -> Event:
    solver = _get_current_solver()
    action = Action(f, convert_to_string(f))
    event = Event(solver, trigger, action)
    return event


# Solving

def solve(t_end: float, time_step: Union[float, None] = 0.1, method='RK45', t_eval: Union[List, NDArray] = None,
          include_crossing_times: bool = True, plot: bool = True, rtol: float = 1e-3,
          atol: float = 1e-6, max_step=np.inf, verbose: bool = False) -> None:
    """
    Solve the equations of the dynamical system through a hybrid solver.

    The hybrid solver is a modified version of SciPy's solve_ivp() function.

    :param max_step: Maximum allowed step size. Default is np.inf, i.e., the step size is not bounded and determined
        solely by the solver.
    :param plot: If True, a plot will show the result of the simulation for variables that were registered to plot.
    :param verbose: If True, print solving information to the console.
    :param include_crossing_times: If True, include time points at which events are triggered.
    :param t_end: Time at which the integration stops.
    :param method: Integration method to use. Default is 'RK45'. For a list of available methods, see SciPy's
        `solve_ivp()` documentation.
    :param time_step: Time step for the integration. If None, use points selected by the solver.
    :param t_eval: Times at which to store the computed solution. If None, use points selected by the solver.
    :param rtol: Relative tolerance. The solver keeps the local error estimates less than `atol + rtol * abs(y)`.
    :param atol: Absolute tolerance. The solver keeps the local error estimates less than `atol + rtol * abs(y)`.
    """
    solver = _get_current_solver()
    solver.solve(t_end, method, time_step, t_eval, include_crossing_times=include_crossing_times, verbose=verbose,
                 plot=plot, rtol=rtol, atol=atol, max_step=max_step)


def explore(fun: Callable[..., T], t_end: float, bounds=(), time_step: float = None, title: str = "") -> None:
    """
    Explore the function f over the given bounds and solve the system until t_end.
    This function needs the sliderplot package.

    :param title: Title of the plot
    :param time_step: Time step of the simulation
    :param fun: The function to explore.
    :param t_end: Time at which the integration stops.
    :param bounds: Bounds for the exploration.
    """
    solver = _get_current_solver()
    solver.explore(fun, t_end, bounds, time_step, title)


def new_system() -> None:
    """
    Create a new solver system.
    """
    new_solver = Solver()
    _solver_list.append(new_solver)


AVAILABLE_EXPORT_FILE_FORMATS = ["csv", "json"]


def export_to_df(*variables: TemporalVar) -> "pd.DataFrame":
    import pandas as pd

    solver = _get_current_solver()
    if not solver.solved:
        raise Exception("System must be solved before exporting the results. Please call 'vip.solve(t_end)'.")
    variables_dict = {"Time (s)": solver.t}
    if not variables:
        variable_dict = {**variables_dict, **{key: var.values for key, var in solver.named_vars.items()}}
    else:
        variable_dict = {**variables_dict, **{get_expression(var): var.values for var in variables}}
    variable_dict["Time (s)"] = solver.t
    return pd.DataFrame(variable_dict)


def export_file(filename: str, variable_list: Iterable[TemporalVar] = None,
                file_format: Literal["csv", "json"] = None) -> None:
    if file_format is None:
        file_format = Path(filename).suffix.lstrip(".")
    if file_format not in AVAILABLE_EXPORT_FILE_FORMATS:
        raise ValueError(
            f"Unsupported file format: {file_format}. "
            f"The available file formats are {', '.join(AVAILABLE_EXPORT_FILE_FORMATS)}"
        )
    df = export_to_df(*variable_list)
    if file_format == "csv":
        df.to_csv(filename, index=False)
    elif file_format == "json":
        df.to_json(filename, orient="records")


def clear() -> None:
    """
    Clear the current solver's stored state.
    Use it to free memory before creating a new simulation.
    """
    solver = _get_current_solver()
    solver.clear()


### I do not like the current implementation, so I disable it for the moment.
#
# def save(*args: TemporalVar) -> None:
#     """
#     Save the given TemporalVars with their variable names.
#
#     :param args: TemporalVars to be saved.
#     :raises ValueError: If any of the arguments is not a TemporalVar.
#     """
#     solver = _get_current_solver()
#     if not all([isinstance(arg, TemporalVar) for arg in args]):
#         raise ValueError("Only TemporalVars can be saved.")
#     for i, variable in enumerate(args):
#         variable_name = argname(f'args[{i}]')
#         solver.saved_vars[variable_name] = variable


def get_var(var_name: str) -> TemporalVar:
    """
    Retrieve a saved TemporalVar by its name.

    :param var_name: The name of the saved TemporalVar.
    :return: The retrieved TemporalVar.
    """
    solver = _get_current_solver()
    return solver.saved_vars[var_name]


def plot() -> None:
    """
    Plot the variables that have been marked for plotting.
    """
    solver = _get_current_solver()
    solver.plot()


def _get_current_solver() -> "Solver":
    if not _solver_list:
        new_system()
    return _solver_list[-1]


def _check_solver_discrepancy(input_value: Union["TemporalVar", float], solver: "Solver") -> None:
    """
    Raise an exception if there is a discrepancy between the input solver and the solver of the input variable.
    :param input_value:
    :param solver:
    """
    if isinstance(input_value, TemporalVar) and not solver is input_value.solver:
        raise Exception("Can not use a variable from a previous system.")


def _convert_to_temporal_var(value: Union[T, TemporalVar[T]]) -> TemporalVar[T]:
    if not isinstance(value, TemporalVar):
        value = temporal(value)
    return value



import pathlib
import runpy
import sys
from typing import Sequence

import matplotlib
import numpy as np
import pytest

import vip_ivp as vip


def test_plot_constant():
    a = vip.temporal(1)
    a.to_plot()
    vip.solve(10)


def test_multiple_loop_into():
    d_n1 = vip.loop_node()
    n1 = vip.integrate(d_n1, 1)
    d_n1.loop_into(-0.3 * n1)
    d_n1.loop_into(-0.2 * n1, force=True)

    d_n2 = vip.loop_node()
    n2 = vip.integrate(d_n2, 1)
    d_n2.loop_into(-0.5 * n2)

    vip.solve(10)
    error_array = n2.values - n1.values
    assert all(error_array < 1e-10)


def test_pendulum():
    dd_th = vip.loop_node()
    d_th = vip.integrate(dd_th, 0)
    th = vip.integrate(d_th, np.pi / 2)
    dd_th.loop_into(-9.81 / 1 * np.sin(th))
    vip.solve(10, time_step=0.1)


def test_source():
    u = vip.temporal(lambda t: 5 * np.sin(5 * t))
    dd_th = vip.loop_node()
    d_th = vip.integrate(dd_th, 0)
    th = vip.integrate(d_th, np.pi / 2)
    dd_th.loop_into(u - 9.81 / 1 * np.sin(th))
    vip.solve(10)


def test_loop():
    acc = vip.loop_node()
    vit = vip.integrate(acc, 0)
    pos = vip.integrate(vit, 5)
    acc.loop_into(0.1 + 1 / 10 * (-1 * vit - 1 * pos) + 5)
    vip.solve(50)


def test_integrate_scalar():
    x = vip.integrate(5, 1)
    vip.solve(10, time_step=1)
    assert np.allclose(x.values, np.linspace(1, 51, 11))


def test_no_integration():
    a = vip.temporal(lambda t: t)
    b = 2 * a

    # a.to_plot('A')
    # b.to_plot('2*A')

    vip.solve(10)


def test_system_without_integration():
    # Without time step
    a = vip.temporal(lambda t: t)
    b = 2 * a
    vip.solve(10)
    assert np.array_equal(2 * a.values, b.values)

    # With time step
    vip.new_system()
    a = vip.temporal(lambda t: t)
    b = 2 * a
    vip.solve(10, time_step=0.1)
    assert np.array_equal(2 * a.values, b.values)


def test_plant_controller():
    def controller(error):
        ki = 1
        kp = 1
        i_err = vip.integrate(ki * error, x0=0)
        return i_err + kp * error

    def plant(x):
        m = 1
        k = 1
        c = 1
        v0 = 0
        x0 = 5
        acc = vip.loop_node()
        vit = vip.integrate(acc, v0)
        pos = vip.integrate(vit, x0)
        acc.loop_into(1 / m * (x - c * vit - k * pos + x))
        return pos

    target = 1
    error = vip.loop_node()
    x = controller(error)
    y = plant(x)
    error.loop_into(target - y)

    vip.solve(50)


def test_mass_spring_bond_graph():
    def inertia(forces: Sequence[vip.TemporalVar], mass: float):
        acc = sum(forces) / mass + 9.81
        vit = vip.integrate(acc, 0)
        return vit

    def spring(speed1, speed2, stiffness: float):
        x = vip.integrate(speed1 - speed2, 0)
        force2 = k * x
        force1 = -force2
        return force1, force2

    k = 1
    mass = 1
    speed1 = vip.loop_node()
    force1, force2 = spring(speed1, 0, k)
    vit = inertia((force1,), mass)
    speed1.loop_into(vit)

    vip.solve(50)


def test_differentiate():
    d_n = vip.loop_node()
    n = vip.integrate(d_n, 1)
    d_n.loop_into(-0.5 * n)
    d_n2 = n.derivative()

    d_n2.to_plot()
    d_n.to_plot()

    vip.solve(10, time_step=0.01)

    print(d_n.values)
    print(d_n2.values)

    errors = d_n.values - d_n2.values
    assert all(errors[1:] < 0.01)


def test_integrated_differentiation():
    step = vip.temporal(lambda t: 0 if t < 1 else 1)
    # Differentiate then integrate
    d_step_bad = step.derivative()
    step_bad = vip.integrate(d_step_bad, 0)

    # Integrate then differentiate
    i_step = vip.integrate(step, 0)
    step_ok = i_step.derivative()

    step.to_plot()
    step_ok.to_plot()
    # d_step_bad.to_plot()
    step_bad.to_plot()

    vip.solve(2, time_step=0.01)


def test_float_crossing_event():
    a = vip.temporal(lambda t: t)

    crossing = a.crosses(5)
    vip.terminate_on(crossing)

    vip.solve(10, time_step=1)
    print(a.values)
    print(a.t)
    assert len(a.t) == 6
    assert a.values[-1] == 5


def test_boolean_crossing_event():
    a = vip.temporal(lambda t: t)
    cond = a >= 5

    crossing = cond.crosses(True)
    vip.terminate_on(crossing)

    vip.solve(10, time_step=1)
    print(cond.values)
    print(cond.t)
    assert len(a.t) == 6
    assert cond.values[-1] == True


def test_string_crossing_event():
    a = vip.temporal(lambda t: t)
    string = vip.where(a >= 5, "Aa", "Ba")

    crossing = string.crosses("Aa")
    vip.terminate_on(crossing)

    vip.solve(10, time_step=1)
    print(string.values)
    print(string.t)
    assert len(a.t) == 6
    assert string.values[-1] == "Aa"


def test_bouncing_projectile_motion():
    # Parameters
    GRAVITY = -9.81
    v0 = 20
    th0 = np.radians(45)
    mu = 0.1  # Coefficient of air drag

    # Compute initial condition
    v0 = [v0 * np.cos(th0), v0 * np.sin(th0)]
    x0 = [0, 0]

    k = 0.7  # Bouncing coefficients
    v_min = 2

    # Create system
    acceleration = vip.loop_node(2)
    velocity = vip.integrate(acceleration, v0)
    position = vip.integrate(velocity, x0)
    v_norm = np.sqrt(velocity[0] ** 2 + velocity[1] ** 2)
    acceleration.loop_into([-mu * velocity[0] * v_norm,
                            GRAVITY - mu * velocity[1] * v_norm])

    hit_ground = position[1].crosses(0, "falling")
    bounce = velocity[1].reset_on(hit_ground & (abs(velocity[1]) > v_min), -k * velocity[1])
    vip.terminate_on(hit_ground & (abs(velocity[1]) <= v_min))

    position.to_plot("Position")
    velocity[1].to_plot()

    vip.solve(20, time_step=0.2, verbose=True)
    print(position.t)


def test_eval_events_at_all_time_points():
    # Parameters
    GRAVITY = -9.81
    v0 = 20
    th0 = np.radians(45)
    mu = 0.1  # Coefficient of air drag

    # Compute initial condition
    v0 = [v0 * np.cos(th0), v0 * np.sin(th0)]
    x0 = [0, 0]

    k = 0.7  # Bouncing coefficients
    v_min = 0.01

    # Create system
    acceleration = vip.loop_node(2)
    velocity = vip.integrate(acceleration, v0)
    position = vip.integrate(velocity, x0)
    v_norm = np.sqrt(velocity[0] ** 2 + velocity[1] ** 2)
    acceleration.loop_into([-mu * velocity[0] * v_norm,
                            GRAVITY - mu * velocity[1] * v_norm])

    stopped = abs(velocity[1]) < v_min

    hit_ground = position[1].crosses(
        0,
        direction="falling"
    )

    velocity[1].reset_on(hit_ground, -k * velocity[1]),

    vip.terminate_on(stopped)

    # position.to_plot("Position")
    stopped.to_plot("Stopping condition")

    vip.solve(20, time_step=0.01)
    # print(position.t)
    assert np.count_nonzero(stopped.values) == 1


def test_eval_events_at_all_time_points_with_trigger():
    # Parameters
    GRAVITY = -9.81
    v0 = 20
    th0 = np.radians(45)
    mu = 0.1  # Coefficient of air drag

    # Compute initial condition
    v0 = [v0 * np.cos(th0), v0 * np.sin(th0)]
    x0 = [0, 0]

    k = 0.7  # Bouncing coefficients
    v_min = 0.01

    # Create system
    acceleration = vip.loop_node(2)
    velocity = vip.integrate(acceleration, v0)
    position = vip.integrate(velocity, x0)
    v_norm = np.sqrt(velocity[0] ** 2 + velocity[1] ** 2)
    acceleration.loop_into([-mu * velocity[0] * v_norm,
                            GRAVITY - mu * velocity[1] * v_norm])

    stopped = abs(velocity[1]) < v_min

    hit_ground = position[1].crosses(
        0,
        direction="falling"
    )

    velocity[1].reset_on(hit_ground, -k * velocity[1]),

    stop_trigger = stopped.crosses(True)
    vip.terminate_on(stop_trigger)

    # position.to_plot("Position")
    stopped.to_plot("Stopping condition")
    stop_trigger.to_plot()

    vip.solve(20, time_step=0.01)
    # print(position.t)
    assert np.count_nonzero(stopped.values) == 1


def test_multiple_events_at_the_same_instant():
    a = vip.temporal(1)
    ia = vip.integrate(a, 0)

    inhibit = vip.integrate(0, 1)

    t1 = vip.interval_trigger(2)
    t2 = vip.timeout_trigger(6)
    e1 = ia.reset_on(t1 & inhibit, 0)
    e2 = inhibit.reset_on(t2, 0)

    ia.to_plot()

    vip.solve(10, time_step=0.01)

    assert ia.values[-1] == 4


def test_demos():
    matplotlib.use("Agg")
    demo_dir = pathlib.Path(__file__).parent.parent / "demos"

    # Get all .py files in demo folder
    demo_scripts = list(demo_dir.glob("*.py"))
    demo_scripts = [path for path in demo_scripts if "explore" not in str(path)]
    print(demo_scripts)

    for script_path in demo_scripts:
        vip.new_system()
        print(script_path)
        try:
            runpy.run_path(str(script_path), run_name="__main__")
        except Exception as e:
            pytest.fail(f"Demo script {script_path.name} raised an exception: {e}")


def test_forgiving_temporal_functions():
    """
    Test if the temporal function can accept functions that do not support array inputs
    """

    def non_vec_fun(t):
        return max(1.0 - 0.005 * t, 0)

    a = vip.temporal(non_vec_fun)

    a.to_plot()
    vip.solve(10)


def test_forgiving_f():
    def non_vec_fun(x):
        return max(x, 1)

    # Test f
    a = vip.temporal(lambda t: t)
    b = vip.f(non_vec_fun)(a)

    a.to_plot()
    b.to_plot()

    vip.solve(10)


def test_loads_of_recursion():
    a = vip.loop_node()
    b = a.delayed(10)
    a.loop_into(b + 1)

    a.to_plot()
    b.to_plot()

    vip.solve(1000, plot=False, time_step=0.05)

    print(a.values)


def test_big_delay():
    a = vip.loop_node()
    b = a.delayed(100)
    a.loop_into(b + 1)

    a.to_plot()
    b.to_plot()

    vip.solve(1000, time_step=0.05)

    print(a.values)


# def test_cascading_events():
#     # Parameters
#     initial_height = 1  # m
#     GRAVITY = -9.81
#     k = 0.7  # Bouncing coefficient
#     v_min = 0.01  # Minimum velocity need to bounce
#
#     # Create the system
#     acceleration = vip.temporal(GRAVITY)
#     velocity = vip.integrate(acceleration, x0=0)
#     height = vip.integrate(velocity, x0=initial_height)
#
#     count = vip.temporal(0)
#
#     # Create the bouncing event
#     bounce = vip.where(abs(velocity) > v_min, velocity.action_reset_to(-k * velocity), vip.action_terminate)
#     height.on_crossing(0, bounce, terminal=False, direction="falling")
#     velocity.on_crossing(0, count.action_set_to(count + 1), direction="rising")
#
#     # Add variables to plot
#     height.to_plot("Height (m)")
#     velocity.to_plot()
#     count.to_plot()
#
#     # Solve the system
#     vip.solve(20, time_step=0.001)
#
#     assert count.values[-1] == 18


def test_stiff_ode():
    dy = vip.loop_node(3)
    # Robertson problem
    y = vip.integrate(dy, [1, 0, 0])
    dy1 = -0.04 * y[0] + 1e4 * y[1] * y[2]
    dy2 = 0.04 * y[0] - 1e4 * y[1] * y[2] - 3e7 * y[1] ** 2
    dy3 = 3e7 * y[1] ** 2
    dy.loop_into([dy1, dy2, dy3])

    vip.solve(1e2, method="BDF")
    print(y.values)



import pytest

import vip_ivp as vip


def test_algebraic_loop():
    x = vip.loop_node()
    ix = vip.integrate(x, 0)
    x.loop_into(x + ix)

    with pytest.raises(RecursionError):
        vip.solve(10)


def test_loop_node_not_set():
    x = vip.loop_node()
    with pytest.raises(Exception):
        vip.solve(10)

    vip.new_system()
    y = vip.loop_node(shape=(2, 3))
    with pytest.raises(Exception):
        vip.solve(10)

    vip.new_system()
    z=vip.loop_node(shape=(2,3), strict=False)
    vip.solve(10)


def test_set_loop_node_two_times():
    x = vip.loop_node()
    x.loop_into(6)
    with pytest.raises(Exception):
        x.loop_into(5)


def test_crossing_integration_bounds():
    a = vip.temporal(1)
    signal = vip.temporal(lambda t: 6 - t)
    ia = vip.integrate(a, 0, maximum=signal, minimum=-1)

    ia.to_plot("Integral")

    with pytest.raises(ValueError):
        try:
            vip.solve(10, time_step=1)
        except Exception as e:
            print(e)
            raise e

def test_if_statement():
    time = vip.temporal(lambda t: t)
    with pytest.raises(ValueError):
        if time < 5:
            step = vip.temporal(0)
        else:
            step = vip.temporal(1)

        vip.solve(10, time_step=1)

def test_x0_outside_bound():
    with pytest.raises(ValueError):
        x = vip.integrate(5, 0, 2, 10)

def test_integrate_discrete_signal():
    a=vip.temporal(1)
    a_d=a.delayed(1)
    with pytest.raises(NotImplementedError):
        iad=vip.integrate(a_d,0)



import time

import matplotlib
import numpy as np
import pandas as pd

import vip_ivp as vip


def test_collections_get_methods():
    l = vip.temporal([1, 2, 3, 4])
    di = vip.temporal({"a": 5, "b": 6})
    arr = vip.temporal(np.array([1, 2, 3, 4, 5]))
    arr_slice = arr[2:]

    b = vip.integrate(l[0], 0)
    c = vip.integrate(di["b"], 0)
    d = vip.integrate(arr_slice[0], 0)

    vip.solve(10, time_step=1)

    print(l[0].values)
    print(di["b"].values)
    print(arr_slice[0].values)


def test_logical_operators():
    x_a = vip.temporal(lambda t: np.sin(t))
    x_b = vip.temporal(lambda t: np.sin(t + np.pi / 2))
    a = x_a >= 0
    b = x_b >= 0

    # and
    a_and_b = a & b
    b_and_a = b & a
    # or
    a_or_b = a | b
    b_or_a = b | a
    # xor
    a_xor_b = a ^ b
    b_xor_a = b ^ a
    # not
    not_a = ~a

    vip.solve(20)
    # Assertions
    # and
    assert np.array_equal(a_and_b.values, np.logical_and(a.values, b.values))
    assert np.array_equal(b_and_a.values, np.logical_and(a.values, b.values))
    # or
    assert np.array_equal(a_or_b.values, np.logical_or(a.values, b.values))
    assert np.array_equal(b_or_a.values, np.logical_or(a.values, b.values))
    # xor
    assert np.array_equal(a_xor_b.values, np.logical_xor(a.values, b.values))
    assert np.array_equal(b_xor_a.values, np.logical_xor(a.values, b.values))
    # not
    assert np.array_equal(not_a.values, np.logical_not(a.values))


def test_t_eval_over_time_step():
    t_eval = [0, 2, 2.5, 7]
    a = vip.temporal(1)

    vip.solve(10, t_eval=t_eval)

    assert np.array_equal(t_eval, a.t)


def test_solvers():
    methods = ['RK23', 'RK45', 'DOP853', 'Radau', 'BDF', 'LSODA']
    for method in methods:
        vip.new_system()
        d_n = vip.loop_node()
        n = vip.integrate(d_n, 1)
        d_n.loop_into(-0.5 * n)

        start = time.time()
        vip.solve(10, method=method, time_step=0.001)
        print(f"{method}: {time.time() - start}s")


def test_use_numpy_function():
    x = vip.temporal(5)
    a = vip.integrate(x, 0)

    np.random.seed(10)
    map_length = 100
    map_x = np.linspace(0, 100, map_length)
    map_y = np.random.rand(map_length)
    y = vip.f(np.interp)(a, map_x, map_y)

    vip.solve(10, time_step=0.01)

    error_array = y.values - np.interp(a.values, map_x, map_y)

    assert np.all(error_array == 0)


def test_use_basic_function():
    def basic_function(x: float) -> int:
        if x > 0:
            return 1
        else:
            return -1

    input = vip.temporal(lambda t: np.cos(t))
    output = vip.f(basic_function)(input)

    input.to_plot()
    output.to_plot()

    vip.solve(10, plot=False)


def test_use_numpy_method():
    array_source = vip.temporal([lambda t: t, lambda t: 2 * t, lambda t: 3 * t, lambda t: 4 * t])
    reshaped_array = array_source.m(array_source.output_type.reshape)((2, 2))
    square_array_source = vip.temporal([[lambda t: t, lambda t: 2 * t], [lambda t: 3 * t, lambda t: 4 * t]])
    # reshaped_array.to_plot()
    vip.solve(3, 1)
    print(array_source.values[0])
    print(reshaped_array.values)
    print(square_array_source.values)
    # Bug explanation: When the TemporalVariable possess a numpy array that is computed from an operation, it does not
    # manipulate the shape to have the time dimensions the last instead of the first.

    assert np.array_equal(reshaped_array.values, square_array_source.values)


def test_multidimensional_integration_source():
    arr = np.array([5, 4])
    arr_x0 = np.array([1, 0])
    lis = [5, 4]
    lis_x0 = [1, 0]
    dic = {"a": 5, "b": 4}
    dic_x0 = {"a": 1, "b": 0}

    # Integrate with sources
    da = vip.temporal(arr)
    a = vip.integrate(da, arr_x0)
    dd = vip.temporal(dic)
    d = vip.integrate(dd, dic_x0)

    # Integrate with python variables
    a2 = vip.integrate(arr, arr_x0)
    a3 = vip.integrate(lis, lis_x0)
    d2 = vip.integrate(dic, dic_x0)

    vip.solve(10, time_step=1)

    a0_fun = lambda t: 5 * t + 1
    a1_fun = lambda t: 4 * t

    # Get values
    print(a.values)
    print(d.values)

    # Evaluate integration from source
    assert np.allclose(a[0].values, a0_fun(a[0].t))
    assert np.allclose(a[1].values, a1_fun(a[1].t))
    assert np.allclose(a.values[0], a0_fun(a[0].t))
    assert np.allclose(a.values[1], a1_fun(a[1].t))
    assert np.allclose(d["a"].values, a0_fun(a[0].t))
    assert np.allclose(d["b"].values, a1_fun(a[1].t))
    # Evaluate integration from python variables
    assert np.allclose(a2[0].values, a0_fun(a[0].t))
    assert np.allclose(a2[1].values, a1_fun(a[1].t))
    assert np.allclose(a3[0].values, a0_fun(a[0].t))
    assert np.allclose(a3[1].values, a1_fun(a[1].t))
    assert np.allclose(d2["a"].values, a0_fun(a[0].t))
    assert np.allclose(d2["b"].values, a1_fun(a[1].t))


def test_array_integration_loop_node():
    dx = vip.loop_node(shape=2)
    x = vip.integrate(dx, [1, 2])
    dx.loop_into([-0.5 * x[0], -0.4 * x[1]])

    dx1 = vip.loop_node()
    x1 = vip.integrate(dx1, 1.0)
    dx1.loop_into(-0.5 * x1)

    dx2 = vip.loop_node()
    x2 = vip.integrate(dx2, 2.0)
    dx2.loop_into(-0.4 * x2)

    vip.solve(10, time_step=1)

    assert np.array_equal(x[0].values, x1.values)
    assert np.array_equal(x[1].values, x2.values)


def test_multidimensional_integration_loop_node():
    lambdas = np.linspace(0.1, 1, 6).reshape(2, 3)
    d_n = vip.loop_node(shape=(2, 3))
    n = vip.integrate(d_n, np.ones((2, 3)))
    d_n.loop_into(-n * lambdas)

    n.to_plot()

    vip.solve(10, time_step=0.001, plot=False)


def test_multidimensional_differentiation():
    source = [lambda t: t, lambda t: 2 * t, lambda t: 3 * t, lambda t: 4 * t]
    array_source = vip.temporal(source)
    diff = array_source.derivative()
    truth = [array_source[i].derivative() for i in range(len(source))]

    array_source.to_plot()
    diff.to_plot()

    vip.solve(10, time_step=1, plot=False)
    for i in range(len(source)):
        assert np.array_equal(diff[i].values, truth[i].values)


def test_set_loop_node_multiple_times():
    source1 = vip.temporal(lambda t: t)
    source2 = vip.temporal(lambda t: 2 * t)
    loop = vip.loop_node()
    loop.loop_into(source1)
    loop.loop_into(source2, force=True)
    vip.solve(10, time_step=1)
    assert np.allclose(loop.values, np.linspace(0, 30, 11))
    assert loop.expression == "source1 + source2"


def test_conditions():
    a = vip.temporal(lambda t: t)
    b = vip.temporal(5)
    c = vip.where(a < 5, a, b)
    vip.solve(10, time_step=0.1)
    assert np.array_equal(c.values, np.where(a.values < 5, a.values, b.values))


def test_scenario_interpolation():
    scenario_dict = {"t": [0, 1, 2, 3, 4], "a": [1, 2, 3, 4, 5], "b": [0, 10, -10, 10, -10]}
    scenario_df = pd.DataFrame(scenario_dict)
    scenario_json = "tests/files/scenarii/scenario_basic.json"
    scenario_csv = "tests/files/scenarii/scenario_basic.csv"

    scenarii_inputs = [scenario_df, scenario_dict, scenario_json, scenario_csv]

    for scenario in scenarii_inputs:
        print(f"Test scenario: {scenario}")
        vip.new_system()
        scenario_variable = vip.create_scenario(scenario, time_key="t", sep=";", interpolation_kind="linear")

        vip.solve(4, time_step=0.5)

        a = scenario_variable["a"]
        b = scenario_variable["b"]

        assert a.values[0] == 1
        assert a.values[1] == 1.5
        assert b.values[0] == 0
        assert b.values[1] == 5
        assert b.values[3] == 0


def test_plot_collections():
    arr = np.array([5, 4])
    arr_x0 = np.array([1, 0])
    dic = {"a": 2, "b": 3}
    dic_x0 = {"a": 1, "b": 0}
    arr2 = np.arange(6).reshape(2, 3)

    # Integrate with sources
    da = vip.temporal(arr)
    a = vip.integrate(da, arr_x0)
    dd = vip.temporal(dic)
    d = vip.integrate(dd, dic_x0)
    a2 = vip.temporal(arr2)

    matplotlib.use('Agg')
    a.to_plot("Array")
    a2.to_plot("2D array")
    d.to_plot("Dict")

    vip.solve(10, 0.1)
    print(a2.values)


def test_array_comparisons_operators():
    f1 = lambda t: t
    f2 = lambda t: 2.5 * t

    a = vip.temporal([5, 5])
    b = vip.temporal([f1, f2])

    # Equality
    equ_arr = a == b
    equ1 = b[0] == a[0]
    equ2 = b[1] == a[1]

    # Inequality
    neq_arr = a != b
    neq1 = b[0] != a[0]
    neq2 = b[1] != a[1]

    # Less than
    lt_arr = b < a
    lt1 = b[0] < a[0]
    lt2 = b[1] < a[1]

    # Greater than
    gt_arr = b > a
    gt1 = b[0] > a[0]
    gt2 = b[1] > a[1]

    # Less than or equal
    le_arr = b <= a
    le1 = b[0] <= a[0]
    le2 = b[1] <= a[1]

    # Greater than or equal
    ge_arr = b >= a
    ge1 = b[0] >= a[0]
    ge2 = b[1] >= a[1]

    vip.solve(10, 1)

    # Assertions for equality
    equ_arr.values
    assert np.array_equal(equ_arr[0].values, equ1.values)
    assert np.array_equal(equ_arr[1].values, equ2.values)

    # Assertions for inequality
    assert np.array_equal(neq_arr[0].values, neq1.values)
    assert np.array_equal(neq_arr[1].values, neq2.values)

    # Assertions for less than
    assert np.array_equal(lt_arr[0].values, lt1.values)
    assert np.array_equal(lt_arr[1].values, lt2.values)

    # Assertions for greater than
    assert np.array_equal(gt_arr[0].values, gt1.values)
    assert np.array_equal(gt_arr[1].values, gt2.values)

    # Assertions for less than or equal
    assert np.array_equal(le_arr[0].values, le1.values)
    assert np.array_equal(le_arr[1].values, le2.values)

    # Assertions for greater than or equal
    assert np.array_equal(ge_arr[0].values, ge1.values)
    assert np.array_equal(ge_arr[1].values, ge2.values)


def test_bounded_integration_by_constant():
    a = vip.temporal(1)
    ia_inc = vip.integrate(a, 2, maximum=5, minimum=2)
    ia_dec = vip.integrate(-a, 5, maximum=5, minimum=2)

    # ia_inc.to_plot("Integral")
    # ia_dec.to_plot("Decreasing integral")

    vip.solve(10)

    assert ia_inc.values[-1] == 5
    assert ia_inc.values[0] == 2
    assert ia_dec.values[0] == 5
    assert ia_dec.values[-1] == 2


def test_bounded_integration_by_variable():
    a = vip.temporal(1)
    signal = vip.temporal(lambda t: 6 - t)
    ia_inc = vip.integrate(a, 0, maximum=signal)
    ia_dec = vip.integrate(-a, 0, minimum=-signal)

    # ia_inc.to_plot("Integral")
    # ia_dec.to_plot("Decreasing integral")

    vip.solve(10, time_step=1)

    assert ia_inc.values[3] == 3
    assert ia_inc.values[-1] == -4
    assert ia_dec.values[3] == -3
    assert ia_dec.values[-1] == 4


def test_delete_event():
    a = vip.temporal(lambda t: t)

    crossing = a.crosses(6)
    vip.terminate_on(crossing & (a < 3))

    vip.solve(10)
    print(a.solver.events)

    assert a.t[-1] == 10


def test_variable_step_solving():
    # Exponential decay : dN/dt = - λ * N
    d_n = vip.loop_node()
    n = vip.integrate(d_n, 1)
    d_n.loop_into(-0.5 * n)

    # Choose which variables to plot
    # n.to_plot()
    # d_n.to_plot()

    # Solve the system. The plot will automatically show.
    vip.solve(10, time_step=None)


def test_simultaneous_actions():
    a = vip.temporal(5)
    ia = vip.integrate(a, 0)

    crossing = ia.crosses(10)
    ia.reset_on(crossing, 0)
    vip.terminate_on(crossing)

    ia.to_plot("IA")

    vip.solve(10)

    assert ia.t[-1] == 2


def test_set_timeout():
    a = vip.temporal(1)
    ia = vip.integrate(a, 0)

    timeout = vip.timeout_trigger(3)
    ia.reset_on(timeout, 0)

    ia.to_plot()
    timeout.to_plot()

    vip.solve(10, time_step=1)
    assert ia.values[3] == 0


def test_set_interval():
    a = vip.temporal(1)
    ia = vip.integrate(a, 0)

    interval = vip.interval_trigger(2)
    ia.reset_on(interval, 0)

    ia.to_plot()
    interval.to_plot()

    vip.solve(10, time_step=1)

    assert np.all(ia.values[0::2] == 0)
    assert np.allclose(ia.values[1::2], np.full_like(ia.values[1::2], 1))


# TEMPORARILY DELETED. IT WILL BECOME RELEVANT AGAIN WITH STATEFUL VARIABLES

# def test_increment_timeout():
#     count = vip.temporal(0)
#
#     vip.set_timeout(count.action_set_to(count + 1), 2)
#
#     count.to_plot()
#     vip.solve(10, time_step=1)
#
#     assert count.values[0] == 0
#     assert count.values[2] == 1
#     assert count.values[-1] == 1
#
#
# def test_increment_interval():
#     count = vip.temporal(0)
#     vip.set_interval(count.action_set_to(count + 1), 2)
#
#     count.to_plot()
#     vip.solve(10, time_step=1)
#
#     print(count.values)
#     print(count.t)
#
#     assert count.values[0] == 0
#     assert count.values[2] == 1
#     assert count.values[4] == 2
#     assert count.values[6] == 3
#     assert count.values[8] == 4
#     assert count.values[10] == 5


def test_loop_with_delay():
    start = time.time()
    a = vip.loop_node()
    b = a.delayed(1)

    a.loop_into(b + 1)

    # a.to_plot()
    # b.to_plot()

    vip.solve(10, time_step=1)
    print(a.values)
    print(time.time() - start)


def test_state_variable():
    def relay(u, previous_state):
        u_up = 0.5
        u_low = -0.5
        if u > u_up:
            state = 1
        elif u < u_low:
            state = -1
        else:
            state = previous_state
        y = 5 * state
        return y, state

    u = vip.temporal(lambda t: np.sin(t))
    relay_state = vip.loop_node()
    previous_state = relay_state.delayed(1, 1)
    o = vip.f(relay)(u, previous_state)[:]
    y = o[0]
    relay_state.loop_into(o[1])

    u.to_plot()
    relay_state.to_plot()

    vip.solve(10)



import numpy as np

import vip_ivp as vip
from scipy.integrate import solve_ivp


def rc_circuit_vip(q0=1, r=1, c=1):
    vip.new_system()
    dq = vip.loop_node()
    q = vip.integrate(dq, q0)
    dq.loop_into(-q / (r * c))
    vip.solve(10, time_step=0.001)
    return q.values


def rc_circuit_scipy(q0=1, r=1, c=1):
    # r * dq/dt + q/c = 0
    t_eval = np.linspace(0, 10, 10001)

    def dy(t, y):
        return -y[0] / (c * r)

    sol = solve_ivp(dy, [0, 10], [q0], t_eval=t_eval)
    return sol.y[0]


def stiff_ode_vip():
    vip.new_system()
    dy=vip.loop_node(3)
    # Robertson problem
    y=vip.integrate(dy,[1,0,0])
    dy1 = -0.04 * y[0] + 1e4 * y[1] * y[2]
    dy2 = 0.04 * y[0] - 1e4 * y[1] * y[2] - 3e7 * y[1] ** 2
    dy3 = 3e7 * y[1] ** 2
    dy.loop_into([dy1,dy2,dy3])

    vip.solve(1e3, method="BDF")
    return y.values


def stiff_ode_scipy():
    def robertson(t, y):
        dy1 = -0.04 * y[0] + 1e4 * y[1] * y[2]
        dy2 = 0.04 * y[0] - 1e4 * y[1] * y[2] - 3e7 * y[1] ** 2
        dy3 = 3e7 * y[1] ** 2
        return [dy1, dy2, dy3]

    y0 = [1, 0, 0]
    t_span = (0, 1e3)  # Long time span
    t_eval=np.linspace(0,t_span[1],10001)

    # Use a stiff solver like Radau or BDF
    sol = solve_ivp(robertson, t_span, y0, method='BDF', t_eval=t_eval)
    return sol.y


def test_differential_equation_equality():
    assert np.allclose(rc_circuit_vip(), rc_circuit_scipy())


def test_differential_equation_vip(benchmark):
    result = benchmark(rc_circuit_vip)


def test_differential_equation_scipy(benchmark):
    result = benchmark(rc_circuit_scipy)

def test_stiff_ode_equality():
    assert np.allclose(stiff_ode_vip(), stiff_ode_scipy())

def test_stiff_ode_vip(benchmark):
    result = benchmark(stiff_ode_vip)


def test_stiff_ode_scipy(benchmark):
    result = benchmark(stiff_ode_scipy)



from .api import temporal, integrate, loop_node, solve, explore, clear, new_system, \
    get_var, plot, f, TemporalVar, Solver, LoopNode, create_scenario, export_file, export_to_df, \
    where, terminate_on, timeout_trigger, interval_trigger, execute_on



from typing import TYPE_CHECKING

import numpy as np
from scipy.integrate._ivp.bdf import BDF
from scipy.integrate._ivp.radau import Radau
from scipy.integrate._ivp.rk import RK23, RK45, DOP853
from scipy.integrate._ivp.lsoda import LSODA
from scipy.optimize import OptimizeResult
from scipy.integrate._ivp.common import EPS, OdeSolution
from scipy.integrate._ivp.base import OdeSolver

if TYPE_CHECKING:
    from src.vip_ivp.base import Event, CrossTriggerVar

METHODS = {'RK23': RK23,
           'RK45': RK45,
           'DOP853': DOP853,
           'Radau': Radau,
           'BDF': BDF,
           'LSODA': LSODA}

MESSAGES = {0: "The solver successfully reached the end of the integration interval.",
            1: "A termination event occurred."}


class OdeResult(OptimizeResult):
    pass


def solve_event_equation(cross_var: "CrossTriggerVar", sol, t_low, t_up, is_discrete: bool = False,
                         crossing_tolerance=1e-12):
    """Solve an equation corresponding to an ODE event.

    The equation is ``event(t, y(t)) = 0``, here ``y(t)`` is known from an
    ODE solver using some sort of interpolation. It is solved by
    `scipy.optimize.brentq` with xtol=atol=4*EPS.

    Parameters
    ----------
    event : callable
        Function ``event(t, y)``.
    sol : callable
        Function ``sol(t)`` which evaluates an ODE solution between `t_old`
        and  `t`.
    t_old, t : float
        Previous and new values of time. They will be used as a bracketing
        interval.

    Returns
    -------
    root : float
        Found solution.
    """
    from scipy.optimize import brentq
    if is_discrete:
        return t_up
    else:
        if abs(cross_var.function(t_low, sol(t_low))) <= crossing_tolerance:
            return t_low
        elif abs(cross_var.function(t_up, sol(t_up))) <= crossing_tolerance:
            return t_up
        else:
            return brentq(lambda t: cross_var.function(t, sol(t)), t_low, t_up,
                          xtol=4 * EPS, rtol=4 * EPS)


def is_discrete(cross_trigger: "CrossTriggerVar") -> bool:
    return cross_trigger.function.output_type in (str, bool, np.bool)


def handle_events(sol, events, active_events_indices, t_old, t, t_eval):
    """Helper function to handle events.

    Parameters
    ----------
    sol : DenseOutput
        Function ``sol(t)`` which evaluates an ODE solution between `t_old`
        and  `t`.
    events : list of callables, length n_events
        Event functions with signatures ``event(t, y)``.
    active_events : ndarray
        Indices of events which occurred.
    event_count : ndarray
        Current number of occurrences for each event.
    max_events : ndarray, shape (n_events,)
        Number of occurrences allowed for each event before integration
        termination is issued.
    t_old, t : float
        Previous and new values of time.

    Returns
    -------
    root_indices : ndarray
        Indices of events which take zero between `t_old` and `t` and before
        a possible termination.
    roots : ndarray
        Values of t at which events occurred.
    terminate : bool
        Whether a terminal event occurred.
    """
    active_events = [events[idx] for idx in active_events_indices]
    roots = [solve_event_equation(e, sol, t_old, t, is_discrete(e), t_eval)
             for e in active_events]
    roots = np.asarray(roots)
    return active_events, roots


def find_active_events(events, sol, t_eval, t, t_old):
    """Find which event occurred during an integration step.

    Parameters
    ----------
    g, g_new : array_like, shape (n_events,)
        Values of event functions at a current and next points.
    direction : ndarray, shape (n_events,)
        Event "direction" according to the definition in `solve_ivp`.

    Returns
    -------
    active_events : ndarray
        Indices of events which occurred during the step.
    """

    g = [e.g for e in events]
    direction = np.array([e.direction for e in events])
    t_upper = t
    t_lower = t_old

    if t_eval is None:
        t_list = []
    else:
        t_eval_i_new = np.searchsorted(t_eval, t, side="right")
        t_eval_step = t_eval[:t_eval_i_new]
        t_list = t_eval_step[t_eval_step > t_old]
    # Prevent events that triggered at the previous step to trigger again in this step, because its g_new is at 0 so
    # an irrelevant zero-crossing is sure to occur.
    previous_triggers_mask = np.array([not t_old in e.t_events for e in events])

    t_list = [*t_list, t]
    for i, t_ev in enumerate(t_list):
        g_new = [e(t_ev, sol(t_ev)) for e in events]
        active_events_indices = find_active_events_in_step(g, g_new, direction, previous_triggers_mask)
        if active_events_indices.size > 0:
            t_upper = t_ev
            return active_events_indices, t_upper, t_lower
        if i == 0 and len(t_list) > 1:
            # Disable the preventing of zero-crossing from previously triggered events
            g = g_new
            t_lower = t_ev
            previous_triggers_mask = None
    return np.array([]), t_upper, t_lower


def find_active_events_in_step(g, g_new, direction, previous_triggers_mask=None, crossing_tolerance=1e-12):
    """Find which event occurred during an integration step.

    Parameters
    ----------
    g, g_new : array_like, shape (n_events,)
        Values of event functions at a current and next points.
    direction : ndarray, shape (n_events,)
        Event "direction" according to the definition in `solve_ivp`.

    Returns
    -------
    active_events : ndarray
        Indices of events which occurred during the step.
    """

    # replace False values by -1 because False being equal 0 breaks the 0 crossing detection.
    g = [x if not isinstance(x, (bool, np.bool)) or x == True else -1 for x in g]
    g_new = [x if not isinstance(x, (bool, np.bool)) or x == True else -1 for x in g_new]

    g, g_new, direction = np.asarray(g), np.asarray(g_new), np.asarray(direction)
    up = (g <= crossing_tolerance) & (g_new >= -crossing_tolerance)
    down = (g >= -crossing_tolerance) & (g_new <= crossing_tolerance)
    either = up | down
    mask = (up & (direction > 0) |
            down & (direction < 0) |
            either & (direction == 0))
    if previous_triggers_mask is not None:
        mask = mask & previous_triggers_mask
    active_events_indices = np.nonzero(mask)[0]
    return active_events_indices



import numpy as np

import vip_ivp as vip
from matplotlib import pyplot as plt

ABSOLUTE_TOLERANCE = 0.01


def test_t_end():
    a = vip.temporal(lambda t: t)
    vip.solve(10, 2)
    assert a.t[-1] == 10


def test_rc_circuit():
    # r * dq/dt + q/c = 0
    q0_values = np.linspace(1, 10, 5)
    r_values = np.linspace(1, 10, 5)
    c_values = np.linspace(1, 10, 5)
    t_end = 100

    for q0 in q0_values:
        for R in r_values:
            for C in c_values:
                # Compute exact solution
                t = np.linspace(0, t_end, 1001)
                exact_solution = q0 * np.exp(-t / (R * C))
                # Compute solver solution
                vip.new_system()
                dq = vip.loop_node()
                q = vip.integrate(dq, q0)
                dq.loop_into(-q / (R * C))
                vip.solve(t_end, t_eval=t)
                error_array = exact_solution - q.values
                assert all(error_array < ABSOLUTE_TOLERANCE)


def test_harmonic_equation():
    # y'' + 9 * y = 0
    # Compute exact solution
    x_end = 10
    x = np.linspace(0, x_end, 11)
    y_exact = np.cos(3 * x) + 2 / 3 * np.sin(3 * x)
    # Compute solver solution
    ddy = vip.loop_node()
    dy = vip.integrate(ddy, 2)
    y = vip.integrate(dy, 1)
    ddy.loop_into(-9 * y)
    vip.solve(x_end, t_eval=x)
    error_array = y_exact - y.values
    assert all(error_array < ABSOLUTE_TOLERANCE)


def test_second_order_ode():
    # y'' + 4 * y' + 4 * y = 0
    # Compute exact solution
    x_end = 100
    x = np.linspace(0, x_end, 1001)
    y_exact = (2 * x + 1) * np.exp(-2 * x)
    # Compute solver solution
    ddy = vip.loop_node()
    dy = vip.integrate(ddy, 0)
    y = vip.integrate(dy, 1)
    ddy.loop_into(-4 * dy - 4 * y)
    vip.solve(x_end, t_eval=x)
    error_array = y_exact - y.values
    assert all(error_array < ABSOLUTE_TOLERANCE)


def test_bouncing_ball():
    gravity = -9.81
    h = 10
    k = 0.5
    v_min = 0.1

    time_step = 0.01

    def dy(t, v0):
        return gravity * t + v0

    def y(t, v0, h0):
        return 0.5 * gravity * t ** 2 + v0 * t + h0

    def t_ground(v0, h0):
        return (-v0 - np.sqrt(v0 ** 2 - 2 * gravity * h0)) / gravity

    t = np.arange(0, 10 + time_step / 2, time_step)
    solution = np.zeros_like(t)
    t0 = 0
    v0 = 0
    current_h = h

    while True:
        t_g = t0 + t_ground(v0, current_h)
        current_sol = y(t - t0, v0, current_h)
        solution[(t0 <= t) & (t < t_g)] = current_sol[(t0 <= t) & (t < t_g)]
        v0 = dy(t_g - t0, v0)
        print(f"{t_g=} {v0=}")
        if abs(v0) < v_min:
            solution = solution[t <= t_g]
            t = t[t <= t_g]
            break
        v0 = -k * v0
        t0 = t_g
        current_h = 0

    acc = vip.temporal(gravity)
    velocity = vip.integrate(acc, 0)
    h = vip.integrate(velocity, h)

    hit_ground = h.crosses(0, "falling")
    velocity.reset_on(hit_ground, -k * velocity)
    vip.terminate_on(hit_ground & (0 > velocity) & (velocity >= -v_min))

    vip.solve(10, time_step=time_step, plot=False, include_crossing_times=False)

    # plt.plot(t, solution)
    # plt.plot(h.t, h.values)
    # plt.plot(velocity.t, velocity.values)
    # plt.plot(hit_ground.t, hit_ground.values)
    # plt.hlines([-v_min, v_min], 0, 5)
    # plt.grid()
    # plt.show()

    assert np.allclose(h.values, solution)
